###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This is the workflow 'corehole' using the Fleur code, which calculates binding
energies and corelevel shifts with different methods.
'divide and conquer'
"""

# TODO maybe also calculate the reference structure to check on the supercell calculation
# TODO creation of wf_para nodes for scf fleurinp_changes for corelevel has to be right
# TODO maybe always rewrite hole econfig tag in inp.xml otherwise it might lead to errors,
# be careful with LOs.
# TODO corelevel workflow, rename species of 0,0,0 position in inp.xml

#import os.path

import re
import numpy as np
from pprint import pprint
from aiida.plugins import DataFactory
from aiida.orm import Code, load_node, CalcJobNode
from aiida.orm import Int, StructureData, Dict, RemoteData
from aiida.engine import WorkChain, if_, ToContext
from aiida.engine import submit
#from aiida.work.process_registry import ProcessRegistry
from aiida.engine.processes.functions import calcfunction as cf
from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.tools.StructureData_util import supercell
from aiida_fleur.tools.create_corehole import create_corehole_para  #, create_corehole_fleurinp
from aiida_fleur.tools.extract_corelevels import extract_corelevels
from aiida_fleur.tools.StructureData_util import break_symmetry
from aiida_fleur.tools.StructureData_util import find_equi_atoms
from aiida_fleur.tools.element_econfig_list import get_econfig, get_coreconfig
from aiida_fleur.tools.element_econfig_list import econfigstr_hole, states_spin
from aiida_fleur.tools.element_econfig_list import get_state_occ, highest_unocc_valence
from aiida_fleur.tools.dict_util import dict_merger, extract_elementpara
from aiida_fleur.data.fleurinp import FleurinpData


class FleurCoreholeWorkChain(WorkChain):
    """
    Turn key solution for a corehole calculation with the FLEUR code.
    Has different protocols for different core-hole types (valence, charge).

    Calculates supercells. Extracts binding energies
    for certain corelevels from the total energy differences a the calculation with
    corehole and without.

    Documentation:
    See help for details.

    Two paths are possible:

    (1) Start from a structure -> workchains run inpgen first (recommended)
    (2) Start from a Fleurinp data object

    Also it is recommended to provide a calc parameter node for the structure

    :param wf_parameters: Dict node, specify, resources and what should be calculated
    :param structure: structureData node, crystal structure
    :param calc_parameters: Dict node, inpgen parameters for the crystal structure
    :param fleurinp:  fleurinpData node,
    :param inpgen: Code node,
    :param fleur: Code node,

    :return: output_corehole_wc_para Dict node,  successful=True if no error

    :uses workchains: fleur_scf_wc, fleur_relax_wc
    :uses calcfunctions: supercell, create_corehole_result_node, prepare_struc_corehole_wf

    """
    # This block of commented code was removed from the docstring and should be put
    # to the other place in the documentation.
    # minimum input example:
    # 1. Code1, Code2, Structure, (Parameters), (wf_parameters)
    # 2. Code2, FleurinpData, (wf_parameters)

    # maximum input example:
    # 1. Code1, Code2, Structure, Parameters,
    #    wf_parameters: {
    #         'method' : 'valence', # what method to use, default for valence to highest open shell
    #         'hole_charge' : 1.0,       # what is the charge of the corehole? 0<1.0
    #         'atoms' : ['all'],           # coreholes on what atoms, positions or index for list, or element ['Be', (0.0, 0.5, 0.334), 3]
    #         'corelevel': ['all'],        # coreholes on which corelevels [ 'Be1s', 'W4f', 'Oall'...]
    #         'supercell_size' : [2,1,1], # size of the supercell [nx,ny,nz]
    #         'para_group' : None,       # use parameter nodes from a parameter group
    #         #'references' : 'calculate',# at some point aiida will have fast forwarding
    #         'relax' : False,          # relax the unit cell first?
    #         'relax_mode': 'Fleur',    # what releaxation do you want
    #         'relax_para' : None, # parameter dict for the relaxation
    #         'scf_para' : None,    # wf parameter dict for the scfs
    #         'same_para' : True,        # enforce the same atom parameter/cutoffs on the corehole calc and ref
    #         'resources' : {"num_machines": 1},# resources per job
    #         'max_wallclock_seconds' : 6*60*60,    # walltime per job
    #         'queue_name' : '',       # what queue to submit to
    #         'serial' : True,           # run fleur in serial, or parallel?
    #         #'job_limit' : 100          # enforce the workflow not to spawn more scfs wcs then this number(which is roughly the number of fleur jobs)
    #         'magnetic' : True          # jspins=2, makes a difference for coreholes
    #         }
    # 2. Code2, FleurinpData, (remote-data), wf_parameters as in 1.

    # Hints:
    # 1. This workflow does not work with local codes!

    _workflowversion = '0.5.1'
    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        },
        'max_wallclock_seconds': 6 * 60 * 60,
        'queue_name': '',
        #'custom_scheduler_commands': '',
        #'import_sys_environment': False,
        #'environment_variables': {}
    }
    _default_wf_para = {
        'method': 'valence',  # what method to use, default for valence to highest open shell
        'hole_charge': 1.0,  # what is the charge of the corehole? 0<1.0
        'atoms':
        ['all'],  # coreholes on what atoms, positions or index for list, or element ['Be', (0.0, 0.5, 0.334), 3]
        'corelevel': ['all'],  # coreholes on which corelevels [ 'Be1s', 'W4f', 'Oall'...]
        'supercell_size': [2, 1, 1],  # size of the supercell [nx,ny,nz]
        'para_group': None,  # use parameter nodes from a parameter group
        #'references' : 'calculate',# at some point aiida will have fast forwarding
        #'relax' : False,          # relax the unit cell first?
        #'relax_mode': 'Fleur',    # what releaxation do you want
        #'relax_para' : None, # parameter dict for the relaxation
        'scf_para': None,  # wf parameter dict for the scfs
        'same_para': True,  # enforce the same atom parameter/cutoffs on the corehole calc and ref
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },  # run fleur in serial, or parallel?
        #'job_limit' : 100          # enforce the workflow not to spawn more scfs wcs then this number(which is roughly the number of fleur jobs)
        'magnetic': True
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('wf_parameters', valid_type=Dict, required=False, default=lambda: Dict(dict=cls._default_wf_para))
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('inpgen', valid_type=Code, required=True)
        spec.input('structure', valid_type=StructureData, required=False)
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.input('options', valid_type=Dict, required=False)  #, default=lambda: Dict(dict=cls._default_options))

        spec.outline(
            cls.check_input,  # first check if input is consistent
            if_(cls.relaxation_needed)(  # ggf relax the given cell
                cls.relax),
            if_(cls.supercell_needed)(  # create a supercell from the given/relaxed cell
                cls.create_supercell),
            cls.create_coreholes,
            cls.run_ref_scf,  # calculate the reference supercell first
            cls.check_scf,
            cls.run_scfs,  # calculate all other corehole calculations
            cls.check_scf,
            cls.return_results)
        spec.output('output_corehole_wc_para', valid_type=Dict)

        spec.exit_code(1, 'ERROR_INVALID_INPUT_RESOURCES', message='The input resources are invalid.')
        spec.exit_code(2, 'ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED', message='Input resources are missing.')
        spec.exit_code(3,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='The code provided is invalid, or not of the right kind.')
        spec.exit_code(4, 'ERROR_INPGEN_CALCULATION_FAILED', message='Inpgen calculation FAILED, check output')
        spec.exit_code(5,
                       'ERROR_CHANGING_FLEURINPUT_FAILED',
                       message='Changing of the FLEURINP data went wrong, check log.')
        spec.exit_code(6,
                       'ERROR_CALCULATION_INVALID_INPUT_FILE',
                       message='The FLEUR input file for the calculation did not validate.')
        spec.exit_code(7,
                       'ERROR_FLEUR_CALCULATION_FAiLED',
                       message='At least one FLEUR calculation FAILED, check the output and log.')
        spec.exit_code(8,
                       'ERROR_CONVERGENCE_NOT_ARCHIVED',
                       message=('At least one FLEUR calculation did not/could not reach the'
                                'desired convergece Criteria, with the current parameters.'))
        spec.exit_code(9,
                       'ERROR_IN_REFERENCE_CREATION',
                       message=('Something went wrong in the determiation what coreholes to '
                                'calculate, probably the input format was not correct. Check log.'))

    def check_input(self):
        """
        init all context parameters, variables.
        Do some input checks. Further input checks are done in further workflow steps
        """
        # TODO: document parameters
        self.report(f'started FleurCoreholeWorkChain version {self._workflowversion}')

        ### init ctx ###

        # internal variables
        self.ctx.calcs_torun = []
        self.ctx.calcs_ref_torun = []
        self.ctx.labels = []
        self.ctx.calcs_res = []

        # input variables
        inputs = self.inputs
        if 'calc_parameters' in inputs:
            self.ctx.ref_para = inputs.get('calc_parameters')
        else:
            self.ctx.ref_para = None

        wf_dict = inputs.wf_parameters.get_dict()
        self.ctx.method = wf_dict.get('method', 'valence')
        self.ctx.joblimit = wf_dict.get('joblimit')
        self.ctx.add_comp_para = wf_dict['add_comp_para']
        self.ctx.same_para = wf_dict.get('same_para')
        self.ctx.scf_para = wf_dict.get('scf_para', {})
        self.ctx.be_to_calc = wf_dict.get('corelevel')
        self.ctx.atoms_to_calc = wf_dict.get('atoms')
        self.ctx.base_structure = inputs.get('structure')  # ggf get from fleurinp
        self.ctx.relax = False
        self.ctx.supercell_size = wf_dict.get('supercell_size', [2, 1, 1])  # 2x2x2 or smaller?
        self.ctx.hole_charge = wf_dict.get('hole_charge', 1.0)
        self.ctx.magnetic = wf_dict.get('magnetic', True)

        defaultoptions = self._default_options
        options = wf_dict.get('options', defaultoptions)
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        #self.ctx.relax = wf_dict.get('relax', default.get('relax'))
        #self.ctx.relax_mode = wf_dict.get('relax_mode', default.get('relax_mode'))
        #self.ctx.relax_para = wf_dict.get('relax_para', default.get('dos_para'))
        self.ctx.base_structure_relax = self.ctx.base_structure

        # return variables initalized here, that at any time an output node can be written.
        self.ctx.successful = True
        self.ctx.bindingenergies = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.hints = []
        self.ctx.cl_energies = []
        self.ctx.all_CLS = []
        self.ctx.ref_cl_energies = []
        self.ctx.fermi_energies = []
        self.ctx.bandgaps = []
        self.ctx.ref_fermi_energies = []
        self.ctx.ref_bandgaps = []
        self.ctx.atomtypes = []
        self.ctx.ref_atomtypes = []
        self.ctx.total_energies = []
        self.ctx.ref_total_energies = []
        self.ctx.wbindingenergies = []
        ### input check ###
        """
        #ususal fleur stuff check
        if fleurinp.get structure
        self.ctx.inputs.base_structure
        wf_para = self.inputs.wf_parameters
        corelevel_to_calc = wf_para.get('corelevel', None)
        if not corelevel_to_calc:
            errormsg = 'You need to specify unter 'corelevel' in the wf_para node on what corelevel you want to have a corehole calculated. (Default is 'all')'
            self.abort_nowait(errormsg)

        """

    def supercell_needed(self):
        """
        check if a supercell is needed and what size it should be
        """
        #think about a rule here to apply 2x2x2 should be enough for nearly everything.
        # but for larger unit cells smaller onces might be ok.
        # So far we just go with what the user has given
        # Is there a way to tell if a supercell was already given as base?
        # Do we want to detect it with some spglib methods?
        self.ctx.supercell_boal = True
        needed = self.ctx.supercell_boal
        # TODO, otherwise in the new system if something else is returned this might fail...?
        return needed

    def create_supercell(self):
        """
        create the needed supercell
        """

        supercell_base = self.ctx.supercell_size
        description = f'WF, Creates a supercell of a crystal structure x{tuple(supercell_base)}.'

        supercell_s = supercell(self.ctx.base_structure_relax,
                                Int(supercell_base[0]),
                                Int(supercell_base[1]),
                                Int(supercell_base[2]),
                                metadata={
                                    'label': 'supercell_wf',
                                    'description': description
                                })

        # overwrite label and description of new structure
        supercell_s.label = f"{'x'.join(map(str,supercell_base))} of {self.ctx.base_structure_relax.uuid}"
        supercell_s.description = supercell_s.description + ' created in a FleurCoreholeWorkChain'
        self.ctx.ref_supercell = supercell_s
        calc_para = self.ctx.ref_para
        if calc_para is None:
            new_calc = supercell_s
        else:
            new_calc = [supercell_s, calc_para]
        self.ctx.calcs_ref_torun.append(new_calc)

    def create_coreholes(self):
        """
        Check the input for the corelevel specification,
        create structure and parameter nodes with all the need coreholes.
        create the wf_parameter nodes for the scfs. Add all calculations to
        scfs_to_run.

        Layout:
        # Check what coreholes should be created.
        # said in the input, look in the original cell
        # These positions are the same for the supercell.
        # break the symmetry for the supercells. (make the corehole atoms its own atom type)
        # create a new species and a corehole for this atom group.
        # move all the atoms in the cell that impurity is in the origin (0.0, 0.0, 0.0)
        # use the fleurinp_change feature of scf to create the corehole after inpgen gen in the scf
        # start the scf with the last charge density of the ref calc? so far no, might not make sense

        # TODO if this becomes to long split
        """
        self.report('INFO: In create_coreholes of FleurCoreholeWorkChain. '
                    'Preparing everything for calculation launches.')

        ########### init variables ##############

        base_struc = self.ctx.base_structure_relax  # one unit cell (given cell)
        base_atoms_sites = base_struc.sites  # list of AiiDA Site types of cell
        base_kinds = base_struc.kinds  # list of AiiDA Kind types of cell
        valid_elements = list(base_struc.get_composition().keys())  # elements in structure
        base_supercell = self.ctx.ref_supercell  # supercell of base cell
        base_k_symbols = {}  #map kind names to elements

        for kind in base_kinds:
            base_k_symbols[kind.name] = kind.symbol

        # we have to find the atoms we want a corelevel on and make them a new kind,
        # also we have to figure out what electron config to set
        atoms_toc = self.ctx.atoms_to_calc  #['Be', (0.0, 0.5, 0.334)/ 3(index)/ 'all']
        if self.ctx.be_to_calc[0] == 'all':
            corelevels_toc_new = []
            for element in valid_elements:
                corelevels_toc_new.append(f'{element}-all')
        else:
            corelevels_toc_new = self.ctx.be_to_calc

        corelevels_toc = corelevels_toc_new  # [ 'Be 1s', 'W_4f', 'O all', 'W-3d'...]

        coreholes_atoms = []  # list of aiida sites
        corehole_to_create = []  # prepare list of dicts for final loop, for calculation creation
        #[{'site' : sites[8], 'kindname' : 'W1', 'econfig': "[Kr] 5s2 4d10 4f13 | 5p6 5d5 6s2", 'fleurinp_change' : []}]

        # get the symmetry equivivalent atoms by ase
        # equi_info_symbol = [['W', 1,2,3,8], ['Be', 4,5,6,7,9] ...]
        #n_equi_info_symbol= {'Be' : count, ...}
        equi_info_symbol, n_equi_info_symbol = find_equi_atoms(base_struc)
        #print(n_equi_info_symbol)
        method = self.ctx.method
        if method == 'valence':
            hole_charge = self.ctx.hole_charge
            correct_val_charge = False  # the routines add the electron per default to the valence
            htype = 'valence'
        elif method == 'charge':
            hole_charge = self.ctx.hole_charge
            correct_val_charge = True
            htype = 'charge'
        else:
            htype = 'valence'  # default so far, otherwise not defined.
            # TODO probally better, to throw error
            hole_charge = self.ctx.hole_charge
            correct_val_charge = False  # the routines add the electron per default to the valence

        ##########
        # 1. Find out what atoms to put coreholes on
        self.report(f'Atoms to calculate : {atoms_toc}')
        for atom_info in atoms_toc:
            if isinstance(atom_info, str):  # , six.text_type)):  #basestring):
                if atom_info == 'all':
                    # add all symmetry equivivalent atoms of structure to create coreholes
                    #coreholes_atoms = base_atoms_sites
                    coreholes_atoms = []
                    for equi_group in equi_info_symbol:
                        # only calculate first element of group, 0 entry is an element string
                        # and there is always a first atom element
                        site_index = equi_group[1][0]
                        coreholes_atoms.append(base_atoms_sites[site_index])
                elif 'all' in atom_info:
                    elem = atom_info.split('all')[0]
                    # check what element we are taking about
                    if elem in valid_elements:
                        for equi_group in equi_info_symbol:
                            # only calculate first element of group, 0 entry is an element string
                            # and there is always a first atom element
                            if equi_group[0] == elem:
                                site_index = equi_group[1][0]
                                coreholes_atoms.append(base_atoms_sites[site_index])
                else:
                    # check if a valid element or some garbage
                    pass
            elif isinstance(atom_info, tuple):  # coordinates
                if len(atom_info) == 3:
                    for site in base_atoms_sites:
                        if site.position == atom_info:  #ggf give a threshold...
                            coreholes_atoms.append(site)
                else:
                    # wrong tuple length this is not a  position
                    self.report(f'WARNING: strange position/coordinates given: {atom_info}')
                    #
            elif isinstance(atom_info, int):  # index for sites
                to_append = None
                try:
                    to_append = base_atoms_sites[atom_info]
                except IndexError:
                    error = (f"ERROR: The index/integer: {atom_info} specified in 'atoms' key is not valid."
                             f'There are only {len(base_atoms_sites)} atom sites in your provided structure.')
                    to_append = None
                    self.report(error)
                if to_append:
                    coreholes_atoms.append(to_append)
            else:
                self.report(f"WARNING: input: {atom_info} of 'atoms' not recongized")

        # TODO: remove doubles in coreholes_atoms?
        #print(coreholes_atoms)
        #print(corelevels_toc)
        dict_corelevel = {}
        # dict_corelevel['W' : {corelevel: ['1s 1/2','4f 7/2', '4f 3/2'], econfig: [config], fleur_changes : []}]

        #########
        # 2. now check what type of corelevel shall we create on those atoms
        self.report(f'Corelevels to calculate : {corelevels_toc}')
        for corel in corelevels_toc:
            if isinstance(corel, str):  # , six.text_type)):  #basestring):
                # split string (Be1s) s.replace(';',' ')... could get rid of re
                elm_cl = re.split('[, ;:-]', corel)
                #print(elm_cl)
                if len(elm_cl) != 2:
                    # something went wrong, wrong input
                    # TODO log, error and hint
                    error = (f'ERROR: corelevel was given in the wrong format: {elm_cl},'
                             'should have len 2. Hint hast to be the format '
                             "['Element,corelevel',...] i.e ['Be,1s', 'W,all]")
                    self.control_end_wc(error)
                    return self.exit_codes.ERROR_IN_REFERENCE_CREATION
                else:
                    # we assume for now ['Element', 'corelevel'] i.e ['Be', '1s']
                    econfigs = []
                    all_corestates = []
                    all_changed_valence = []
                    elm_cl = [str(elm_cl[0]), str(elm_cl[1])]  #otherwise stuff fails because of basestrings
                    if elm_cl[0] in valid_elements:
                        # get corelevel econfig of element
                        dict_corelevel_elm = {}
                        # if econfig given in calc parameter use this econfig...
                        para = self.ctx.ref_para
                        if para is not None:
                            para_dict = para.get_dict()
                            self.report(f'INFO para is here: {para_dict}')
                            element_para = extract_elementpara(para_dict, elm_cl[0])
                            valid_coreconfig = element_para.get('econfig', get_coreconfig(elm_cl[0], full=True))
                        else:
                            valid_coreconfig = get_coreconfig(elm_cl[0], full=True)
                        oriegconfig = get_econfig(elm_cl[0], full=True)
                        highest_unocc = highest_unocc_valence(oriegconfig)
                        if elm_cl[1] == 'all':
                            # add all corelevels to calculate
                            corestates = valid_coreconfig.split()
                            for state in corestates:
                                holeconfig = econfigstr_hole(oriegconfig, state, highest_unocc, htype=htype)
                                rel_states = states_spin.get(state[1], [])
                                for rel in rel_states:
                                    econfigs.append(holeconfig)
                                    all_corestates.append(state + ' ' + rel)
                                    all_changed_valence.append(highest_unocc[:2])
                        elif elm_cl[1] in valid_coreconfig:  # check if corelevel in valid coreconfig
                            #add corelevel to calculate.
                            state_index = oriegconfig.find(elm_cl[1])
                            state = oriegconfig[state_index:state_index + 4].rstrip(' ')  # +4: icii, or ici
                            holeconfig = econfigstr_hole(oriegconfig, state, highest_unocc, htype=htype)
                            rel_states = states_spin.get(state[1], [])
                            # get rel core level (for 4f 5/2, 7/2)
                            for rel in rel_states:
                                econfigs.append(holeconfig)
                                all_corestates.append(state + ' ' + rel)
                                all_changed_valence.append(highest_unocc[:2])  # the methods below need them without occ
                        elif '/' in elm_cl[1]:
                            pass  # TODO FUll state information given...[4f 7/2]
                        else:
                            # corelevel provided wrong, not understood, warning
                            continue
                        # TODO several corelevels of one element... update lists instead of override...
                        dict_corelevel_elm['corelevel'] = all_corestates
                        dict_corelevel_elm['valence'] = all_changed_valence
                        dict_corelevel_elm['econfig'] = econfigs
                        tempd = dict_corelevel.get(elm_cl[0], {})
                        # dict_merger also addes numbers!
                        together = dict_merger(dict_corelevel_elm, tempd)
                        #pprint(together)
                        dict_corelevel[elm_cl[0]] = together
                    else:
                        pass
                        #element or string provieded not in structure,
                        # what about upper and lower caps

        #print(dict_corelevel)
        #output of above
        #list of sites [site_bla, ..]
        #dict_corelevel = {'Be' : {'corelevel' : ['1s1/2'], 'valence' : [], 'econfig' : ['1s2 | 2s2']}}
        # now put atom and corehole information together
        for site in coreholes_atoms:
            selem = base_k_symbols[site.kind_name]
            cl_dict = dict_corelevel.get(selem, None)
            if cl_dict:
                # what coreholes need to be created for that element
                for i, econfig in enumerate(cl_dict.get('econfig', [])):
                    fleurinp_change = []
                    change_kind = site.kind_name + '_corehole1'  # the number at the end
                    # is important otherwise inpgen does not make this a new species
                    # through the hard coded one might lead to conflicts..

                    kind = site.kind_name  # + '1'# this will be the kind name in the broke sym structure,
                    #its name will be changed later to change_kind
                    # maybe the kind name cann also not be known at this point
                    #self.report('{}, {}, {}, {}'.format(econfig, cl_dict.get('corelevel')[i], cl_dict.get('valence')[i], hole_charge))
                    #print(cl_dict.get('corelevel')[i])
                    #print(cl_dict.get('valence')[i])
                    #print(econfig)
                    state_tag_list = get_state_occ(econfig,
                                                   corehole=cl_dict.get('corelevel')[i],
                                                   valence=cl_dict.get('valence')[i],
                                                   ch_occ=hole_charge)
                    attributedict = {'electronConfig': {'stateOccupation': state_tag_list}}
                    #pprint(state_tag_list)
                    change = ('set_species', {
                        'species_name': change_kind,
                        'attributedict': attributedict,
                        'create': False
                    })
                    fleurinp_change.append(change)
                    if correct_val_charge:  # only needed in certain methods
                        charge_change = (
                            'add_number_to_first_attrib',
                            {
                                'attributename': 'valenceElectrons',
                                'add_number': -1.0,  #-hole_charge,  #one electron was added by inpgen, we remove it
                                'mode': 'abs',
                            })
                        fleurinp_change.append(charge_change)
                    elif hole_charge != 1.0:  # fractional valence hole
                        charge_change = (
                            'add_number_to_first_attrib',
                            {
                                'attributename': 'valenceElectrons',
                                'add_number': -1.0 + hole_charge,  # one electron was already added by inpgen
                                'mode': 'abs',
                            })
                        fleurinp_change.append(charge_change)
                    if self.ctx.magnetic:  # Do a collinear magnetic calculation
                        charge_change = ('set_inpchanges', {'change_dict': {'jspins': 2}})
                        fleurinp_change.append(charge_change)
                    #self.report('{}'.format(fleurinp_change))
                    # because there might be already some kinds and another number is right...
                    # repacking of sites, because input to a calcfunction, otherwise not storeable...
                    corehole = {
                        'site': {
                            'kind_name': kind,  #site.kind_name,
                            'position': site.position
                        },
                        'econfig': econfig,
                        'kindname': change_kind,
                        'inpxml_changes': fleurinp_change
                    }
                    corehole_to_create.append(corehole)

        #state_tag_list = get_state_occ(econfigstr, corehole = '', valence = '', ch_occ = 1.0):

        # lesson go over site position to get atom in supercell
        # set econfig for this atom in the supercell
        # (default kind name = element + id) use this for paramter settings

        # fill calcs_torun with (sturcutre, parameter, wf_para)
        #corehole_to_create = [{'site' : sites[8], 'kindname' : 'W1', 'econfig': "[Kr] 5s2 4d10 4f13 | 5p6 5d5 6s2", 'fleurinp_change' : []}]
        calcs = []
        for corehole in corehole_to_create:
            para = self.ctx.ref_para
            wf_para = Dict(dict=corehole)
            #print(corehole)
            #print(base_supercell)
            #print(para)
            # all these steps can be calcfunctions, we have grouped them all in one
            ret_dict = prepare_struc_corehole_wf(base_supercell, wf_para, para)
            moved_struc = ret_dict['moved_struc']
            calc_para = ret_dict['hole_para']
            #print('calc_para:')
            #pprint(calc_para.get_dict())
            #pprint('inpxml_changes {}'.format(corehole['inpxml_changes']))
            # create_wf para or write in last line what should be in 'fleur_change'
            #  for scf, which with the changes in the inp.xml needed
            para = self.ctx.scf_para  # Otherwise inline edit... What about Provenance? TODO check
            if para is None:
                wf_parameter = {}
            else:
                wf_parameter = para
            #print(wf_parameter)
            wf_parameter['add_comp_para'] = self.ctx.add_comp_para
            wf_parameter['inpxml_changes'] = corehole['inpxml_changes']

            wf_parameters = Dict(dict=wf_parameter)
            calcs.append([moved_struc, calc_para, wf_parameters])
        self.ctx.calcs_torun = calcs
        #print('ctx.calcs_torun {}'.format(self.ctx.calcs_torun))
        #self.report('INFO: end of create coreholes')

    '''
    def run_scf2(self):
        """
        Run scf
        """
        calcs = {}
        i = 0
        for scf_input in self.ctx.calcs_ref_torun:
            try:
                res = submit(fleur_scf_wc,
                          fleur=self.inputs.fleur,
                          inpgen = self.inputs.inpgen,
                          **scf_input)
            except: # TODO only if input is wrong
                self.report('WARNING: something in run_ref_scf which I do not reconise: {}'.format(scf_input))
                continue

            label = str('calc_ref{}'.format(i))
            self.ctx.labels.append(label)
            calcs[label] = res

        self.ctx.calcs_ref_torun = []
        return ToContext(**calcs)#  this is a blocking return
    '''

    def run_ref_scf(self):
        """
        Run a scf for the reference super cell
        """

        # TODO: idea instead of a list, just use a dictionary...
        self.report('INFO: In run_ref_scf FleurCoreholeWorkChain')
        print('INFO: In run_ref_scf FleurCoreholeWorkChain')
        para = self.ctx.scf_para
        if para is None:
            wf_parameter = {}
        else:
            wf_parameter = para
        wf_parameter['add_comp_para'] = self.ctx.add_comp_para
        wf_parameters = Dict(dict=wf_parameter)
        options = Dict(dict=self.ctx.options)
        '''
        #res_all = []
        calcs = {}

        i = 0
        for node in self.ctx.calcs_ref_torun: # usually just 1, but we leave the default.
            #print(node)
            i = i+1
            if isinstance(node, StructureData):
                res = fleur_scf_wc.run(wf_parameters=wf_parameters, structure=node,
                            inpgen = self.inputs.inpgen, fleur=self.inputs.fleur)#
            elif isinstance(node, FleurinpData):
                res = fleur_scf_wc.run(wf_parameters=wf_parameters, structure=node,
                            inpgen = self.inputs.inpgen, fleur=self.inputs.fleur)#
            elif isinstance(node, tuple):
                if isinstance(node[0], StructureData) and isinstance(node[1], Dict):
                    #print(node[1].get_dict())
                    res = fleur_scf_wc.run(wf_parameters=wf_parameters, calc_parameters=node[1], structure=node[0],
                                inpgen = self.inputs.inpgen, fleur=self.inputs.fleur)#
                else:
                    self.report(' WARNING: a tuple in run_ref_scf which I do not reconise: {}'.format(node))
            else:
                self.report('WARNING: something in run_ref_scf which I do not reconise: {}'.format(node))
                continue

            #calc_node = res['output_scf_wc_para'].get_inputs()[0] # if run is used, otherwise use labels
            label = str('calc_ref{}'.format(i))
            self.ctx.labels.append(label)
            calcs[label] = res
            #res_all.append(res)
            #self.ctx.calcs_res.append(res)

        '''
        #res_all = []
        calcs = {}
        scf_label = 'FleurCoreholeWorkChain ref cell'
        scf_desc = '|FleurCoreholeWorkChain|'
        i = 0
        for node in self.ctx.calcs_ref_torun:  # usually just 1, but we leave the default.
            #print node
            i = i + 1
            if isinstance(node, StructureData):
                res = self.submit(FleurScfWorkChain,
                                  wf_parameters=wf_parameters,
                                  structure=node,
                                  inpgen=self.inputs.inpgen,
                                  fleur=self.inputs.fleur,
                                  options=options,
                                  metadata={
                                      'label': scf_label,
                                      'description': scf_desc
                                  })  #
            elif isinstance(node, FleurinpData):
                res = self.submit(FleurScfWorkChain,
                                  wf_parameters=wf_parameters,
                                  structure=node,
                                  inpgen=self.inputs.inpgen,
                                  fleur=self.inputs.fleur,
                                  options=options,
                                  metadata={
                                      'label': scf_label,
                                      'description': scf_desc
                                  })  #
            elif isinstance(node, list):
                if isinstance(node[0], StructureData) and isinstance(node[1], Dict):
                    res = self.submit(FleurScfWorkChain,
                                      wf_parameters=wf_parameters,
                                      options=options,
                                      calc_parameters=node[1],
                                      structure=node[0],
                                      inpgen=self.inputs.inpgen,
                                      fleur=self.inputs.fleur,
                                      metadata={
                                          'label': scf_label,
                                          'description': scf_desc
                                      })  #
                else:
                    self.report(f'WARNING: a tuple in run_ref_scf which I do not reconise: {node}')
            else:
                self.report(f'WARNING: something in run_ref_scf which I do not reconise: {node}')
                continue

            #calc_node = res['output_scf_wc_para'].get_inputs()[0] # if run is used, otherwise use labels
            label = str(f'calc_ref{i}')
            self.ctx.labels.append(label)
            calcs[label] = res
            #res_all.append(res)
            #self.ctx.calcs_res.append(res)

        self.ctx.calcs_ref_torun = []
        return ToContext(**calcs)  #  this is a blocking return

    def check_scf(self):
        """
        Check if ref scf was successful, or something needs to be dealt with.
        If unsuccesful abort, because makes no sense to continue.
        """
        #so far not implemented

        for i, label in enumerate(self.ctx.labels):
            calc = self.ctx[label]
            print('collect results ...')
            print(calc)
            if not calc.is_finished_ok:
                self.report(f'SCF workchain {calc} failed')
                print(f'SCF workchain {calc} failed')

    def relaxation_needed(self):
        """
        If the structures should be relaxed, check if their Forces are below a certain
        threshold, otherwise throw them in the relaxation wf.
        """
        self.report('In relaxation FleurCoreholeWorkChain')
        if self.ctx.relax:
            # TODO check all forces of calculations
            forces_fine = True
            return bool(forces_fine)
        else:
            return False

    def relax(self):
        """
        Do structural relaxation for certain structures.
        """
        self.report('In relax FleurCoreholeWorkChain workflow')
        self.ctx.base_structure_relax = self.ctx.base_structure
        #for calc in self.ctx.dos_to_calc:
        #    pass
        #    # TODO run relax workflow

    def run_scfs(self):
        """
        Run a scf for the all corehole calculations in parallel super cell
        """
        self.report('INFO: In run_scfs FleurCoreholeWorkChain')
        print('INFO: In run_scfs FleurCoreholeWorkChain')
        para = self.ctx.scf_para
        if para is None:
            wf_parameter = {}
        else:
            wf_parameter = para
        wf_parameter['add_comp_para'] = self.ctx.add_comp_para
        #wf_parameter['queue_name'] = self.ctx.queue
        #wf_parameter['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands
        wf_parameters = Dict(dict=wf_parameter)
        options = Dict(dict=self.ctx.options)
        #res_all = []
        calcs = {}
        scf_label = 'FleurCoreholeWorkChain cell'
        scf_desc = '|FleurCoreholeWorkChain|'
        # now in parallel
        #print self.ctx.ref_calcs_torun
        i = 0  #
        self.report(f'Calculations to launch : {self.ctx.calcs_torun}')
        for node in self.ctx.calcs_torun:
            #print node
            i = i + 1
            if isinstance(node, StructureData):
                res = self.submit(FleurScfWorkChain,
                                  wf_parameters=wf_parameters,
                                  structure=node,
                                  inpgen=self.inputs.inpgen,
                                  fleur=self.inputs.fleur,
                                  options=options,
                                  metadata={
                                      'label': scf_label,
                                      'description': scf_desc
                                  })  #
            elif isinstance(node, FleurinpData):
                res = self.submit(FleurScfWorkChain,
                                  wf_parameters=wf_parameters,
                                  structure=node,
                                  inpgen=self.inputs.inpgen,
                                  fleur=self.inputs.fleur,
                                  options=options,
                                  metadata={
                                      'label': scf_label,
                                      'description': scf_desc
                                  })  #
            elif isinstance(node, list):
                if isinstance(node[0], StructureData) and isinstance(node[1], Dict):
                    if isinstance(node[2], Dict):
                        res = self.submit(FleurScfWorkChain,
                                          wf_parameters=node[2],
                                          calc_parameters=node[1],
                                          structure=node[0],
                                          options=options,
                                          inpgen=self.inputs.inpgen,
                                          fleur=self.inputs.fleur,
                                          metadata={
                                              'label': scf_label,
                                              'description': scf_desc
                                          })  #
            else:
                self.report(f'ERROR: Something in run_scfs which I do not recognize: {node}')
                continue
            label = str(f'calc{i}')
            #print(label)
            #calc_node = res['output_scf_wc_para'].get_inputs()[0] # if run is used, otherwise use labels
            self.ctx.labels.append(label)
            calcs[label] = res
            #res_all.append(res)
            #print res
            #self.ctx.calcs_res.append(res)
            #self.ctx.calcs_torun.remove(node)
            #print res
            self.to_context(**{label: res})
        self.ctx.calcs_torun = []
        #return ToContext(**calcs)

    def collect_results(self):
        """
        Collect results from certain calculation, check if everything is fine,
        calculate the wanted quantities. currently all energies are in hartree (as provided by Fleur)
        """

        # TODO: what about partial collection?
        # if some calc failed do not abort, but collect the others.
        message = ('INFO: Collecting results of FleurCoreholeWorkChain workflow')
        self.report(message)

        all_CLS = {}
        ref_calcs = []
        ref_cl_energies = {}
        cl_energies = {}
        bindingenergies = []  # atomtype, binding eneergy
        weighted_binding_energies = []  # *(1+1- coreholecharge)
        calcs = []
        # get results from calc/scf
        #calcs = self.ctx.calcs_res
        for i, label in enumerate(self.ctx.labels):
            calc = self.ctx[label]
            #print('collect results ...')
            #print(calc)
            #if not calc.is_finished_ok:
            #    print('calculation failed')
            #    continue
            if i == 0:
                ref_calcs.append(calc)
            else:
                calcs.append(calc)

        fermi_energies, bandgaps, atomtypes, all_corelevel, total_energies = extract_results_corehole(calcs)
        ref_fermi_energies, ref_bandgaps, ref_atomtypes, ref_all_corelevel, ref_total_energies = extract_results_corehole(
            ref_calcs)

        # now calculate binding energies of the coreholes.
        # Differences of total energies
        #for number, energy in total_energies:#.iteritems():
        for energy in total_energies:  #.iteritems():
            #print ref_total_energies
            bde = energy - ref_total_energies[0]  #.get('0', 0)
            bindingenergies.append(bde)
            hole_charge = self.ctx.hole_charge
            if hole_charge != 0.0:
                weighted_binding_energy = bde * (1.0 / hole_charge)
            else:
                weighted_binding_energy = bde
            weighted_binding_energies.append(weighted_binding_energy)
        # make a return dict
        self.ctx.cl_energies = cl_energies
        self.ctx.all_CLS = all_CLS
        self.ctx.ref_cl_energies = ref_cl_energies
        self.ctx.fermi_energies = fermi_energies
        self.ctx.bandgaps = bandgaps
        self.ctx.ref_fermi_energies = ref_fermi_energies
        self.ctx.ref_bandgaps = ref_bandgaps
        self.ctx.atomtypes = atomtypes
        self.ctx.ref_atomtypes = ref_atomtypes
        self.ctx.total_energies = total_energies
        self.ctx.ref_total_energies = ref_total_energies
        self.ctx.bindingenergies = bindingenergies
        self.ctx.wbindingenergies = weighted_binding_energies
        #print(bindingenergies)
        #print(weighted_binding_energies)
        #return

    def return_results(self):
        """
        return the results of the calculations
        """
        # TODO: make sure ouputnodes are always produced
        # get
        # TODO: Maybe all variables should come from the context, therefore they
        # they will be proper initialiezed and you can call return_results, on a controlled
        # abort of the wc. with all output nodes produced....

        #print('coreholes were calculated bla bla')
        # call one routine, that will set all variables in the ctx
        #cl, cls, ref_cl, efermi, gap, ref_efermi, ref_gap, at, at_ref, te, te_ref = self.collect_results()
        # check if this should be called
        if self.ctx.successful:  # TODO parse partially results...
            self.collect_results()

        outputnode_dict = {}

        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['total_energy_ref'] = self.ctx.ref_total_energies
        outputnode_dict['total_energy_ref_units'] = 'eV'
        outputnode_dict['total_energy_all'] = self.ctx.total_energies
        outputnode_dict['total_energy_all_units'] = 'eV'
        outputnode_dict['binding_energy'] = self.ctx.bindingenergies
        outputnode_dict['binding_energy_units'] = 'eV'
        outputnode_dict['weighted_binding_energy'] = self.ctx.wbindingenergies  # BE for scaled hole charge 1.0
        outputnode_dict['weighted_binding_energy_units'] = 'eV'
        outputnode_dict['binding_energy_convention'] = 'negativ'
        outputnode_dict['corehole_type'] = self.ctx.method
        outputnode_dict['coreholes_calculated'] = ''  # on what atom what level basicly description of the other lists
        outputnode_dict['coreholes_calculated_details'] = ''  # the dict internally used
        #outputnode_dict['corelevel_energies'] = cl
        #outputnode_dict['reference_corelevel_energies'] = ref_cl
        outputnode_dict['fermi_energy'] = self.ctx.fermi_energies
        outputnode_dict['fermi_energy_unit'] = 'eV'
        outputnode_dict['coresetup'] = []  #cls
        outputnode_dict['reference_coresetup'] = []  #cls
        outputnode_dict['bandgap'] = self.ctx.bandgaps
        outputnode_dict['bandgap_units'] = 'eV'
        outputnode_dict['reference_bandgaps'] = self.ctx.ref_bandgaps
        outputnode_dict['atomtypes'] = self.ctx.atomtypes
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['errors'] = self.ctx.errors
        outputnode_dict['hints'] = self.ctx.hints

        outputnode = Dict(dict=outputnode_dict)
        outdict = {}
        outdict['output_corehole_wc_para'] = outputnode

        # To have to ouput node linked to the calculation output nodes
        outnodedict = {}
        outnode = Dict(dict=outputnode_dict)
        outnodedict['results_node'] = outnode

        # TODO: bad design, make bullet proof.
        for i, label in enumerate(self.ctx.labels):
            calc = self.ctx[label]
            #print(calc)
            #print(calc.get_outgoing().all())
            try:
                calc_dict = calc.get_outgoing().get_node_by_label(
                    'output_scf_wc_para')  #calc.outputs.output_scf_wc_para
            except (KeyError, ValueError):
                print('continue 2')
            outnodedict[label] = calc_dict

        outdict = create_corehole_result_node(**outnodedict)

        #outdict = {}
        #outdict['output_eos_wc_para'] = ouputnode

        for k, v in outdict.items():
            self.out(k, v)
        msg = ('INFO: FleurCoreholeWorkChain workflow Done')
        self.report(msg)

    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain.
        report errors and always initalize/produce output nodes.
        But log successful=False
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.ctx.errors.append(errormsg)
        self.report(errormsg)
        self.return_results()


@cf
def create_corehole_result_node(**kwargs):  #*args):
    """
    This is a pseudo wf, to create the rigth graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_corehole_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice then have a circle in the database for now...
    #output_para = args[0]
    #return {'output_eos_wc_para'}
    return outdict


#                    corehole = {'site' : {'kind_name' : change_kind,#site.kind_name,
#                                          'position' : site.position},
#                               'econfig' : econfig, 'kindname' : kind,
#                               'inpxml_changes' : fleurinp_change}
@cf
def prepare_struc_corehole_wf(
    base_supercell,
    wf_para,
    para=None,
):  #, _label='prepare_struc_corehole_wf', _description='WF, used in the corehole_wc, breaks the symmetry and moves the cell, prepares the inpgen parameters for a corehole.'):
    """
    calcfunction which does all/some the structure+calcparameter manipulations together
    (therefore less nodes are produced and proverance is kept)
    wf_para: Dict node dict: {'site' : sites[8], 'kindname' : 'W1', 'econfig': "[Kr] 5s2 4d10 4f13 | 5p6 5d5 6s2", 'fleurinp_change' : []}
    """
    from aiida_fleur.tools.StructureData_util import move_atoms_incell

    #from aiida.orm.data.structure import Site

    wf_para_dict = wf_para.get_dict()
    # has to be repacked, site object is not jason serializable...
    site_info = wf_para_dict['site']
    #site = Site(kind_name=site_info['kind_name'], position=site_info['position'])
    pos = site_info['position']  #site.position
    species_name = wf_para_dict['kindname']
    broke_kn = site_info['kind_name']
    new_kinds_names = {broke_kn: [species_name]}
    #print pos
    npos = -np.array(pos)

    # break the symmetry, make corehole atoms its own species. # pos has to be tuple, unpack problem here.. #TODO rather not so nice
    inputs = dict(structure=base_supercell,
                  atoms=[],
                  site=[],
                  pos=[(pos[0], pos[1], pos[2])],
                  new_kinds_names=new_kinds_names)
    if para is not None:
        inputs['parameterdata'] = para
    new_struc, new_para = break_symmetry(**inputs)
    #kinds = new_struc.kinds
    #for kind in kinds:
    #    if kind.name == broke_kn:
    #        kind.name = species_name
    #        print('kindset?')
    # move unit cell that impurity is in 0,0,0
    moved_struc = move_atoms_incell(new_struc, npos)
    # Make sure to provide a parameter node otherwise create_corhole para won't work
    para = create_corehole_para(
        moved_struc,
        species_name,  #wf_para_dict['kindname'],
        wf_para_dict['econfig'],
        parameterdata=new_para,
        species_name=species_name)

    # return of a wf has to be dictionary of nodes...
    return {'moved_struc': moved_struc, 'hole_para': para}


def extract_results_corehole(calcs):
    """
    Collect results from certain calculation, check if everything is fine,
    calculate the wanted quantities.

    params: calcs : list of scf workchains nodes
    """
    # TODO maybe import from somewhere move to common wf

    calc_uuids = []
    for calc in calcs:
        print(calc)
        print(calc.exit_status, calc.exit_message)
        print(calc.get_outgoing().all())
        try:
            calc_uuid = calc.outputs.output_scf_wc_para.get_dict()['last_calc_uuid']
        except (KeyError, AttributeError):
            print('continue')
            continue
        if calc_uuid is not None:
            calc_uuids.append(calc_uuid)
    #print(calc_uuids)

    #all_corelevels = {}
    #fermi_energies = {}
    #bandgaps = {}
    #all_atomtypes = {}
    #all_total_energies = {}

    all_corelevels = []
    fermi_energies = []
    bandgaps = []
    all_atomtypes = []
    all_total_energies = []
    # more structures way: divide into this calc and reference calcs.
    # currently the order in calcs is given, but this might change if you submit
    # check if calculation pks belong to successful fleur calculations
    for i, uuid in enumerate(calc_uuids):
        calc = load_node(uuid)
        if not isinstance(calc, CalcJobNode):
            #raise ValueError("Calculation with pk {} must be a FleurCalculation".format(pk))
            # log and continue
            continue
        if calc.is_finished_ok:
            # get out.xml file of calculation
            #outxml = calc.outputs.retrieved.folder.get_abs_path('path/out.xml')
            outxml = calc.outputs.retrieved.open('out.xml')
            #print outxml
            try:
                corelevels, atomtypes = extract_corelevels(outxml)
            finally:
                outxml.close()
            #all_corelevels.append(core)
            #print('corelevels: {}'.format(corelevels))
            #print('atomtypes: {}'.format(atomtypes))
            #for i in range(0,len(corelevels[0][0]['corestates'])):
            #    print corelevels[0][0]['corestates'][i]['energy']

            #TODO how to store?
            efermi = calc.res.fermi_energy
            #print efermi
            bandgap = calc.res.bandgap
            total_energy = calc.res.energy
            total_energy_units = calc.res.energy_units
            #total_energy = calc.res.total_energy
            #total_energy_units = calc.res.total_energy_units

            # TODO: maybe different, because it is prob know from before
            #fleurinp = calc.inp.fleurinpdata
            #structure = fleurinp.get_structuredata(fleurinp)
            #compound = structure.get_formula()
            #print compound
            #number = '{}'.format(i)
            #fermi_energies[number] = efermi
            #bandgaps[number] = bandgap
            #all_atomtypes[number] = atomtypes
            #all_corelevels[number] = corelevels
            #all_total_energies[number] = total_energy
        else:
            # log and continue
            total_energy = 2e308  #float('nan'))
            bandgap = 2e308  #float('nan')
            efermi = 2e308  #float('nan')
            corelevels = [2e308]  #[float('nan')]
            atomtypes = [2e308]  #[float('nan')]
            #continue
            #raise ValueError("Calculation with pk {} must be in state FINISHED".format(pk))
        fermi_energies.append(efermi)
        bandgaps.append(bandgap)
        all_atomtypes.append(atomtypes)
        all_corelevels.append(corelevels)
        all_total_energies.append(total_energy)

    return fermi_energies, bandgaps, all_atomtypes, all_corelevels, all_total_energies



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module you find the workchain 'FleurScfWorkChain' for the self-consistency
cycle management of a FLEUR calculation with AiiDA.
"""
# TODO: more info in output, log warnings
# TODO: make smarter, ggf delete mixing_history or restart with more or less iterations
# you can use the pattern of the density convergence for this
# TODO: maybe write dict schema for wf_parameter inputs, how?
from lxml import etree

from aiida.orm import Code, load_node
from aiida.orm import StructureData, RemoteData, Dict, Bool, Float
from aiida.engine import WorkChain, while_, if_, ToContext
from aiida.engine import calcfunction as cf
from aiida.common.exceptions import NotExistent

from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur, get_inputs_inpgen
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf import find_last_submitted_calcjob
from aiida_fleur.tools.create_kpoints_from_distance import create_kpoints_from_distance_parameter
from aiida_fleur.workflows.base_fleur import FleurBaseWorkChain

from aiida_fleur.data.fleurinp import FleurinpData, get_fleurinp_from_remote_data

from masci_tools.io.parsers.fleur import outxml_parser


class FleurScfWorkChain(WorkChain):
    """
    Workchain for converging a FLEUR calculation (SCF).

    It converges the charge density, total energy or the largest force.
    Two paths are possible:

    (1) Start from a structure and run the inpgen first optional with calc_parameters
    (2) Start from a Fleur calculation, with optional remoteData

    :param wf_parameters: (Dict), Workchain Specifications
    :param structure: (StructureData), Crystal structure
    :param calc_parameters: (Dict), Inpgen Parameters
    :param fleurinp: (FleurinpData), to start with a Fleur calculation
    :param remote_data: (RemoteData), from a Fleur calculation
    :param inpgen: (Code)
    :param fleur: (Code)

    :return: output_scf_wc_para (Dict), Information of workflow results
        like Success, last result node, list with convergence behavior
    """

    _workflowversion = '0.5.1'
    _default_wf_para = {
        'fleur_runmax': 4,
        'density_converged': 0.00002,
        'energy_converged': 0.002,
        'force_converged': 0.002,
        'kpoints_distance': None,  # in 1/A, usually 0.1
        'kpoints_force_parity': False,
        'kpoints_force_odd': False,
        'kpoints_force_false': False,
        'kpoints_force_gamma': False,
        'nmmp_converged': 0.002,
        'mode': 'density',  # 'density', 'energy' or 'force'
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },
        'itmax_per_run': 30,
        'force_dict': {
            'qfix': 2,
            'forcealpha': 1.0,
            'forcemix': 'straight'
        },
        'use_relax_xml': False,
        'inpxml_changes': [],
        'straight_iterations': None,
        'initial_straight_mixing': False,
        'initial_ldau_straight_mixing': False,
        'initial_ldau_straight_mix_param': 0.0,  #Density matrix frozen by default, since it is the most stable option
    }

    _default_options = {
        'optimize_resources': True,
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 6 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('inpgen', valid_type=Code, required=False)
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('structure', valid_type=StructureData, required=False)
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('remote_data', valid_type=RemoteData, required=False)
        spec.input('options', valid_type=Dict, required=False)
        spec.input('settings', valid_type=Dict, required=False)
        spec.input('settings_inpgen', valid_type=Dict, required=False)
        spec.outline(cls.start, cls.validate_input,
                     if_(cls.fleurinpgen_needed)(cls.run_fleurinpgen), cls.run_fleur, cls.inspect_fleur, cls.get_res,
                     while_(cls.condition)(cls.run_fleur, cls.inspect_fleur, cls.get_res), cls.return_results)

        spec.output('fleurinp', valid_type=FleurinpData)
        spec.output('output_scf_wc_para', valid_type=Dict)
        spec.output('last_fleur_calc_output', valid_type=Dict)
        spec.expose_outputs(FleurBaseWorkChain, namespace='last_calc')

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(233,
                       'ERROR_INVALID_CODE_PROVIDED',
                       message='Input codes do not correspond to fleur or inpgen respectively.')
        spec.exit_code(235, 'ERROR_CHANGING_FLEURINPUT_FAILED', message='Input file modification failed.')
        spec.exit_code(236, 'ERROR_INVALID_INPUT_FILE', message="Input file was corrupted after user's modifications.")
        spec.exit_code(360, 'ERROR_INPGEN_CALCULATION_FAILED', message='Inpgen calculation failed.')
        spec.exit_code(361, 'ERROR_FLEUR_CALCULATION_FAILED', message='Fleur calculation failed.')
        spec.exit_code(362, 'ERROR_DID_NOT_CONVERGE', message='SCF cycle did not lead to convergence.')

    def start(self):
        """
        init context and some parameters
        """
        self.report(f'INFO: started convergence workflow version {self._workflowversion}')

        ####### init    #######

        # internal para /control para
        self.ctx.last_base_wc = None
        self.ctx.loop_count = 0
        self.ctx.relax_generated = False
        self.ctx.calcs = []
        self.ctx.abort = False
        self.ctx.reached_conv = True
        self.ctx.run_straight_mixing = False

        wf_default = self._default_wf_para
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        fleur = self.inputs.fleur
        fleur_extras = fleur.extras
        inpgen_extras = None
        if 'inpgen' in self.inputs:
            inpgen = self.inputs.inpgen
            inpgen_extras = inpgen.extras

        defaultoptions = self._default_options.copy()
        user_options = {}
        if 'options' in self.inputs:
            user_options = self.inputs.options.get_dict()
        '''
        # extend options by code defaults given in code extras
        # Maybe do full recursive merge
        if 'queue_defaults' in fleur_extras:
            qd = fleur_extras['queue_defaults']
            queue = user_options.get('queue', 'default')
            defaults_queue = qd.get(queue, {})
            for key, val in defaultoptions.items():
                defaultoptions[key] = defaults_queue.get(key, val)
        '''
        if 'options' in self.inputs:
            options = user_options
        else:
            options = defaultoptions
        # we use the same options for both codes, inpgen resources get overridden
        # and queue does not matter in case of direct scheduler

        # extend options given by user using defaults
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        self.ctx.max_number_runs = self.ctx.wf_dict.get('fleur_runmax', 4)
        self.ctx.description_wf = self.inputs.get('description', '') + '|fleur_scf_wc|'
        self.ctx.label_wf = self.inputs.get('label', 'fleur_scf_wc')
        self.ctx.default_itmax = self.ctx.wf_dict.get('itmax_per_run', 30)
        self.ctx.straight_mixing_iters = self.ctx.wf_dict.get('straight_iterations')
        if self.ctx.straight_mixing_iters is None:
            self.ctx.straight_mixing_iters = self.ctx.default_itmax

        # return para/vars
        self.ctx.successful = True
        self.ctx.parse_last = True
        self.ctx.distance = []
        self.ctx.all_forces = []
        self.ctx.total_energy = []
        self.ctx.nmmp_distance = []
        self.ctx.energydiff = 10000
        self.ctx.forcediff = 10000
        self.ctx.last_charge_density = 10000
        self.ctx.last_nmmp_distance = -10000
        self.ctx.warnings = []
        # "debug": {},
        self.ctx.errors = []
        self.ctx.info = []
        self.ctx.possible_info = [
            'Consider providing more resources',
            'Consider providing a lot more resources',
            'Consider changing the mixing scheme',
        ]
        self.ctx.fleurinp = None
        self.ctx.formula = ''
        self.ctx.total_wall_time = 0

    def validate_input(self):
        """
        # validate input and find out which path (1, or 2) to take
        # return True means run inpgen if false run fleur directly
        """
        extra_keys = []
        for key in self.ctx.wf_dict:
            if key not in self._default_wf_para:
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for SCF contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        inputs = self.inputs
        if 'fleurinp' in inputs:
            self.ctx.run_inpgen = False
            if 'structure' in inputs:
                error = 'ERROR: structure input is not needed because Fleurinp was given'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'inpgen' in inputs:
                error = 'ERROR: inpgen code is not needed input because Fleurinp was given'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'calc_parameters' in inputs:
                error = 'ERROR: calc_parameters input is not needed because Fleurinp was given'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'remote_data' in inputs:
                warning = ('WARNING: Only initial charge density will be copied from the'
                           'given remote folder because fleurinp is given.')
                self.report(warning)
        elif 'structure' in inputs:
            self.ctx.run_inpgen = True
            if not 'inpgen' in inputs:
                error = 'ERROR: StructureData was provided, but no inpgen code was provided'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'remote_data' in inputs:
                warning = ('WARNING: Only initial charge density will be copied from the'
                           'given remote folder because fleurinp is given.')
                self.report(warning)
        elif 'remote_data' in inputs:
            self.ctx.run_inpgen = False
        else:
            error = 'ERROR: No StructureData nor FleurinpData nor RemoteData was provided'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG

        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = 'The code you provided for inpgen of FLEUR does not use the plugin fleur.inpgen'
                self.report(error)
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                return self.exit_codes.ERROR_INVALID_CODE_PROVIDED

        # check the mode in wf_dict
        mode = self.ctx.wf_dict.get('mode')
        if mode not in ['force', 'density', 'energy', 'gw']:
            error = "ERROR: Wrong mode of convergence: one of 'force', 'density', 'energy' or 'gw' was expected."
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        max_iters = self.ctx.wf_dict.get('itmax_per_run')
        if max_iters <= 1:
            error = "ERROR: 'itmax_per_run' should be equal at least 2"
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        straight_iterations = self.ctx.wf_dict.get('straight_iterations')
        if straight_iterations is not None and straight_iterations <= 1:
            error = "ERROR: 'straight_iterations' should be atleast 2 if given"
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        self.ctx.run_straight_mixing = self.ctx.wf_dict.get('initial_straight_mixing') or \
                                       self.ctx.wf_dict.get('initial_ldau_straight_mixing')

        if straight_iterations is not None:
            if not self.ctx.run_straight_mixing:
                error = "ERROR: 'initial_straight_mixing' or 'initial_ldau_straight_mixing' should be True if 'straight_iterations' is given"
                self.report(error)
                return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # check format of inpxml_changes
        fchanges = self.ctx.wf_dict.get('inpxml_changes', [])
        if fchanges:
            for change in fchanges:
                # somehow the tuple type gets destroyed on the way and becomes a list
                if not isinstance(change, (tuple, list)):
                    error = f'ERROR: Wrong Input inpxml_changes wrong format of: {change} should be tuple of 2. I abort'
                    self.report(error)
                    return self.exit_codes.ERROR_INVALID_INPUT_PARAM
        return

    def fleurinpgen_needed(self):
        """
        Returns True if inpgen calculation has to be submitted
        before fleur calculations
        """
        return self.ctx.run_inpgen

    def run_fleurinpgen(self):
        """
        run the inpgen
        """

        ## prepare inputs for inpgen
        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        label = 'scf: inpgen'
        description = f'{self.ctx.description_wf} inpgen on {self.ctx.formula}'

        inpgencode = self.inputs.inpgen

        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None

        if 'settings_inpgen' in self.inputs:
            settings = self.inputs.settings_inpgen
        else:
            settings = None

        # If given kpt_dist has prio over given calc_parameters
        kpt_dist = self.ctx.wf_dict.get('kpoints_distance', None)
        if kpt_dist is not None:
            cf_para_kpt = Dict(
                dict={
                    'distance': kpt_dist,
                    'force_parity': self.ctx.wf_dict.get('kpoints_force_parity', False),
                    'force_even': self.ctx.wf_dict.get('kpoints_force_even', False),
                    'force_odd': self.ctx.wf_dict.get('kpoints_force_odd', False),
                    'include_gamma': self.ctx.wf_dict.get('kpoints_force_gamma', False)
                })
            inputs = {
                'structure': structure,
                'calc_parameters': params,
                'cf_para': cf_para_kpt,
                'metadata': {
                    'call_link_label': 'create_kpoints_from_distance'
                }
            }
            params = create_kpoints_from_distance_parameter(**inputs)

        options = {
            'max_wallclock_seconds': int(self.ctx.options.get('max_wallclock_seconds')),
            'resources': self.ctx.options.get('resources'),
            'queue_name': self.ctx.options.get('queue_name', '')
        }

        inputs_build = get_inputs_inpgen(structure,
                                         inpgencode,
                                         options,
                                         label,
                                         description,
                                         settings=settings,
                                         params=params)

        # Launch inpgen
        self.report('INFO: run inpgen')
        future = self.submit(inputs_build)

        return ToContext(inpgen=future)

    def reset_straight_mixing(self):
        """
        Turn off the straight mixing features again
        """
        if not self.ctx.fleurinp:
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        wf_dict = self.ctx.wf_dict

        fleurmode = FleurinpModifier(self.ctx.fleurinp)

        fleurmode.set_inpchanges({'itmax': self.ctx.default_itmax})
        #Take out straight mixing
        if wf_dict.get('initial_straight_mixing'):
            fleurmode.set_inpchanges({'imix': 'Anderson'})  #TODO: should take the actual value from before
        if wf_dict.get('initial_ldau_straight_mixing'):
            fleurmode.set_inpchanges({'l_linmix': False})

        # validate?
        try:
            fleurmode.show(display=False, validate=True)
        except etree.DocumentInvalid:
            error = ('ERROR: input, user wanted inp.xml changes did not validate')
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_FILE
        except ValueError as exc:
            error = ('ERROR: input, user wanted inp.xml changes could not be applied.'
                     f'The following error was raised {exc}')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        # apply
        out = fleurmode.freeze()
        self.ctx.fleurinp = out
        return

    def change_fleurinp(self):
        """
        This routine sets somethings in the fleurinp file before running a fleur
        calculation.
        """
        self.report('INFO: run change_fleurinp')

        inputs = self.inputs

        # Has to never crash because corresponding check was done in validate function
        if self.ctx.fleurinp:  # something was already changed
            return
        elif 'fleurinp' in inputs:
            fleurin = self.inputs.fleurinp
        elif 'structure' in inputs:
            if not self.ctx['inpgen'].is_finished_ok:
                error = 'Inpgen calculation failed'
                self.control_end_wc(error)
                return self.exit_codes.ERROR_INPGEN_CALCULATION_FAILED
            fleurin = self.ctx['inpgen'].outputs.fleurinpData
        elif 'remote_data' in inputs:
            # In this case only remote_data for input structure is given
            # fleurinp data has to be generated from the remote inp.xml file to use change_fleurinp
            fleurin = get_fleurinp_from_remote_data(self.inputs.remote_data, store=True)
            self.report(
                f'INFO: generated FleurinpData from files {fleurin.files} from remote folder pk={self.inputs.remote_data.pk}'
            )

        wf_dict = self.ctx.wf_dict
        force_dict = wf_dict.get('force_dict')
        converge_mode = wf_dict.get('mode')
        fchanges = wf_dict.get('inpxml_changes', [])

        fleurmode = FleurinpModifier(fleurin)

        itmax = self.ctx.default_itmax
        if self.ctx.run_straight_mixing:
            if self.ctx.loop_count == 0:
                #Set up straight mixing
                itmax = self.ctx.straight_mixing_iters  #Is set further below
                if wf_dict.get('initial_straight_mixing'):
                    fleurmode.set_inpchanges({'imix': 'straight'})
                if wf_dict.get('initial_ldau_straight_mixing'):
                    fleurmode.set_inpchanges({
                        'l_linmix': True,
                        'mixParam': wf_dict.get('initial_ldau_straight_mix_param')
                    })

        # set proper convergence parameters in inp.xml
        if converge_mode == 'density':
            dist = wf_dict.get('density_converged')
            fleurmode.set_inpchanges({'itmax': itmax, 'minDistance': dist})
        elif converge_mode == 'force':
            force_converged = wf_dict.get('force_converged')
            dist = wf_dict.get('density_converged')
            fleurmode.set_inpchanges({
                'itmax': itmax,
                'minDistance': dist,
                'force_converged': force_converged,
                'l_f': True,
                'qfix': force_dict.get('qfix'),
                'forcealpha': force_dict.get('forcealpha'),
                'forcemix': force_dict.get('forcemix')
            })
        elif converge_mode == 'energy':
            dist = 0.0
            fleurmode.set_inpchanges({'itmax': itmax, 'minDistance': dist})

        elif converge_mode == 'gw':
            dist = 0.0
            fleurmode.set_inpchanges({'itmax': itmax, 'minDistance': dist, 'gw': 1})
            if 'settings' in self.inputs:
                self.inputs.settings.append({'additional_retrieve_list': ['basis.hdf', 'pot.hdf', 'ecore']})
                self.inputs.settings.append({'additional_remotecopy_list': ['basis.hdf', 'pot.hdf', 'ecore']})
            else:
                self.inputs.settings = {
                    'additional_retrieve_list': ['basis.hdf', 'pot.hdf', 'ecore'],
                    'additional_remotecopy_list': ['basis.hdf', 'pot.hdf', 'ecore']
                }

        # apply further user dependend changes
        if fchanges:
            try:
                fleurmode.add_task_list(fchanges)
            except (ValueError, TypeError) as exc:
                error = ('ERROR: Changing the inp.xml file failed. Tried to apply inpxml_changes'
                         f', which failed with {exc}. I abort, good luck next time!')
                self.control_end_wc(error)
                return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        # validate?
        try:
            fleurmode.show(display=False, validate=True)
        except etree.DocumentInvalid:
            error = ('ERROR: input, user wanted inp.xml changes did not validate')
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_FILE
        except ValueError as exc:
            error = ('ERROR: input, user wanted inp.xml changes could not be applied.'
                     f'The following error was raised {exc}')
            self.control_end_wc(error)
            return self.exit_codes.ERROR_CHANGING_FLEURINPUT_FAILED

        # apply
        out = fleurmode.freeze()
        self.ctx.fleurinp = out
        return

    def run_fleur(self):
        """
        run a FLEUR calculation
        """
        self.report('INFO: run FLEUR')

        status = self.change_fleurinp()
        if status:
            return status

        if 'settings' in self.inputs:
            settings = self.inputs.settings
        else:
            settings = None

        if self.ctx.run_straight_mixing and self.ctx.loop_count == 1:
            status = self.reset_straight_mixing()
            if status:
                return status

            if settings is None:
                settings = {}
            else:
                settings = settings.get_dict()

            settings.setdefault('remove_from_remotecopy_list', []).append('mixing_history*')

        fleurin = self.ctx.fleurinp

        if self.ctx['last_base_wc']:
            # will this fail if fleur before failed? try needed?
            remote = self.ctx['last_base_wc'].outputs.remote_folder
        elif 'remote_data' in self.inputs:
            remote = self.inputs.remote_data
        else:
            remote = None

        label = ' '
        description = ' '
        if self.ctx.formula:
            label = f'scf: fleur run {self.ctx.loop_count + 1}'
            description = f'{self.ctx.description_wf} fleur run {self.ctx.loop_count + 1} on {self.ctx.formula}'
        else:
            label = f'scf: fleur run {self.ctx.loop_count + 1}'
            description = f'{self.ctx.description_wf} fleur run {self.ctx.loop_count + 1}, fleurinp given'

        code = self.inputs.fleur
        options = self.ctx.options.copy()

        inputs_builder = get_inputs_fleur(code,
                                          remote,
                                          fleurin,
                                          options,
                                          label,
                                          description,
                                          settings,
                                          add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = self.submit(FleurBaseWorkChain, **inputs_builder)
        self.ctx.loop_count = self.ctx.loop_count + 1
        self.report(f'INFO: run FLEUR number: {self.ctx.loop_count}')
        self.ctx.calcs.append(future)

        return ToContext(last_base_wc=future)

    def inspect_fleur(self):
        """
        Analyse the results of the previous Calculation (Fleur or inpgen),
        checking whether it finished successfully or if not, troubleshoot the
        cause and adapt the input parameters accordingly before
        restarting, or abort if unrecoverable error was found
        """

        self.report('INFO: inspect FLEUR')
        try:
            base_wc = self.ctx.last_base_wc
        except AttributeError:
            self.ctx.parse_last = False
            error = 'ERROR: Something went wrong I do not have a last calculation'
            self.control_end_wc(error)
            return self.exit_codes.ERROR_FLEUR_CALCULATION_FAILED

        exit_status = base_wc.exit_status
        if not base_wc.is_finished_ok:
            error = f'ERROR: Last Fleur calculation failed with exit status {exit_status}'
            self.control_end_wc(error)
            return self.exit_codes.ERROR_FLEUR_CALCULATION_FAILED
        else:
            self.ctx.parse_last = True

    def get_res(self):
        """
        Check how the last Fleur calculation went
        Parse some results.
        """
        self.report('INFO: get results FLEUR')

        mode = self.ctx.wf_dict.get('mode')
        if self.ctx.parse_last:
            last_base_wc = self.ctx.last_base_wc
            fleur_calcjob = load_node(find_last_submitted_calcjob(last_base_wc))
            walltime = last_base_wc.outputs.output_parameters.dict.walltime

            if isinstance(walltime, int):
                self.ctx.total_wall_time = self.ctx.total_wall_time + walltime
            with fleur_calcjob.outputs.retrieved.open(fleur_calcjob.process_class._OUTXML_FILE_NAME,
                                                      'rb') as outxmlfile:
                output_dict = outxml_parser(outxmlfile,
                                            minimal_mode=True,
                                            list_return=True,
                                            iteration_to_parse='all',
                                            ignore_validation=True)

            energies = output_dict.get('energy_hartree', [])
            if energies is not None:
                self.ctx.total_energy.extend(energies)

            if 'overall_density_convergence' in output_dict:
                distances = output_dict['overall_density_convergence']
            else:
                distances = output_dict.get('density_convergence', [])

            if distances is not None:
                self.ctx.distance.extend(distances)

            if 'ldau_info' in output_dict:
                nmmp_distances = output_dict['ldau_info'].get('density_matrix_distance', [])

                if nmmp_distances is not None:
                    self.ctx.nmmp_distance.extend(nmmp_distances)

            if mode == 'force':
                forces = output_dict.get('force_atoms', [])
                if forces is not None:
                    for force_iter in forces:
                        self.ctx.all_forces.append([force for atom, force in force_iter])

        else:
            errormsg = 'ERROR: scf wc was not successful, check log for details'
            self.control_end_wc(errormsg)
            return self.exit_codes.ERROR_FLEUR_CALCULATION_FAILED

        if not self.ctx.distance:
            # if fleur relaxes an already converged crystal it stops directly
            if mode == 'force':
                self.report('INFO: System already force converged, could not extract distance.')
                self.ctx.last_charge_density = None
            else:
                errormsg = 'ERROR: did not manage to extract charge density from the calculation'
                self.control_end_wc(errormsg)
                return self.exit_codes.ERROR_FLEUR_CALCULATION_FAILED
        else:
            self.ctx.last_charge_density = self.ctx.distance[-1]

        if self.ctx.nmmp_distance:
            if isinstance(self.ctx.nmmp_distance[-1], list):
                self.ctx.last_nmmp_distance = max(self.ctx.nmmp_distance[-1])
            else:
                self.ctx.last_nmmp_distance = self.ctx.nmmp_distance[-1]

    def condition(self):
        """
        check convergence condition
        """
        self.report('INFO: checking condition FLEUR')
        mode = self.ctx.wf_dict.get('mode')
        ldau_notconverged = False

        energy = self.ctx.total_energy
        if len(energy) >= 2:
            self.ctx.energydiff = abs(energy[-1] - energy[-2])

        if mode == 'force':
            forces = self.ctx.all_forces
            if len(forces) >= 2:
                self.ctx.forcediff = max(
                    abs(forces[-1][i][k] - forces[-2][i][k]) for i in range(len(forces[-1])) for k in range(3))
        else:
            self.ctx.forcediff = 'can not be determined'

        if self.ctx.last_nmmp_distance > 0.0 and \
           self.ctx.last_nmmp_distance >= self.ctx.wf_dict.get('nmmp_converged'):
            ldau_notconverged = True

        if mode == 'density':
            if self.ctx.wf_dict.get('density_converged') >= self.ctx.last_charge_density:
                if not ldau_notconverged:
                    return False
        elif mode in ('energy', 'gw'):
            if self.ctx.wf_dict.get('energy_converged') >= self.ctx.energydiff:
                if not ldau_notconverged:
                    return False
        elif mode == 'force':
            if self.ctx.last_charge_density is None:
                try:
                    _ = self.ctx.last_base_wc.outputs.relax_parameters
                except NotExistent:
                    pass
                else:
                    if not ldau_notconverged:
                        return False

            elif self.ctx.wf_dict.get('density_converged') >= self.ctx.last_charge_density:
                try:
                    _ = self.ctx.last_base_wc.outputs.relax_parameters
                except NotExistent:
                    pass
                else:
                    if not ldau_notconverged:
                        return False

        if self.ctx.loop_count >= self.ctx.max_number_runs:
            self.ctx.reached_conv = False
            return False

        return True

    def return_results(self):
        """
        return the results of the calculations
        This should run through and produce output nodes even if everything failed,
        therefore it only uses results from context.
        """
        if self.ctx.last_base_wc:
            try:
                last_calc_uuid = find_last_submitted_calcjob(self.ctx.last_base_wc)
            except NotExistent:
                last_calc_uuid = None
        else:
            last_calc_uuid = None

        try:  # if something failed, we still might be able to retrieve something
            last_calc_out = self.ctx.last_base_wc.outputs.output_parameters
            retrieved = self.ctx.last_base_wc.outputs.retrieved
            last_calc_out_dict = last_calc_out.get_dict()
        except (NotExistent, AttributeError):
            last_calc_out = None
            last_calc_out_dict = {}
            retrieved = None

        last_nmmp_distance = None
        if self.ctx.last_nmmp_distance > 0.0:
            last_nmmp_distance = self.ctx.last_nmmp_distance

        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['material'] = self.ctx.formula
        outputnode_dict['conv_mode'] = self.ctx.wf_dict.get('mode')
        outputnode_dict['loop_count'] = self.ctx.loop_count
        outputnode_dict['iterations_total'] = last_calc_out_dict.get('number_of_iterations_total', None)
        outputnode_dict['distance_charge'] = self.ctx.last_charge_density
        outputnode_dict['distance_charge_all'] = self.ctx.distance
        outputnode_dict['total_energy'] = last_calc_out_dict.get('energy_hartree', None)
        outputnode_dict['total_energy_all'] = self.ctx.total_energy
        outputnode_dict['force_diff_last'] = self.ctx.forcediff
        outputnode_dict['force_largest'] = last_calc_out_dict.get('force_largest', None)
        outputnode_dict['distance_charge_units'] = 'me/bohr^3'
        outputnode_dict['total_energy_units'] = 'Htr'
        outputnode_dict['nmmp_distance'] = last_nmmp_distance
        outputnode_dict['nmmp_distance_all'] = self.ctx.nmmp_distance
        outputnode_dict['last_calc_uuid'] = last_calc_uuid
        outputnode_dict['total_wall_time'] = self.ctx.total_wall_time
        outputnode_dict['total_wall_time_units'] = 's'
        outputnode_dict['info'] = self.ctx.info
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['errors'] = self.ctx.errors

        num_iterations = last_calc_out_dict.get('number_of_iterations_total', None)
        if self.ctx.successful and self.ctx.reached_conv:
            if len(self.ctx.total_energy) <= 1:  # then len(self.ctx.all_forces) <= 1 too
                self.report('STATUS: Done, the convergence criteria are reached.\n'
                            'INFO: The charge density of the FLEUR calculation '
                            f'converged after {self.ctx.loop_count} FLEUR runs, {num_iterations} '
                            f'iterations and {self.ctx.total_wall_time} sec '
                            f'walltime to {outputnode_dict["distance_charge"]} "me/bohr^3" \n'
                            'INFO: Did not manage to get energy and largest force difference '
                            'between two last iterations, probably converged in a single iteration')
            else:
                self.report('STATUS: Done, the convergence criteria are reached.\n'
                            'INFO: The charge density of the FLEUR calculation '
                            f'converged after {self.ctx.loop_count} FLEUR runs, {num_iterations} '
                            f'iterations and {self.ctx.total_wall_time} sec '
                            f'walltime to {outputnode_dict["distance_charge"]} "me/bohr^3" \n'
                            'INFO: The total energy difference of the last two iterations '
                            f'is {self.ctx.energydiff} Htr and largest force difference is '
                            f'{self.ctx.forcediff} Htr/bohr')
        elif self.ctx.successful and not self.ctx.reached_conv:
            if len(self.ctx.total_energy) <= 1:  # then len(self.ctx.all_forces) <= 1 too
                self.report('STATUS/WARNING: Done, the maximum number of runs '
                            'was reached.\n INFO: The '
                            'charge density of the FLEUR calculation, '
                            f'after {self.ctx.loop_count} FLEUR runs, {num_iterations} '
                            f' iterations and {self.ctx.total_wall_time} sec '
                            f'walltime is {outputnode_dict["distance_charge"]} "me/bohr^3"\n'
                            'INFO: can not extract energy and largest force difference between'
                            ' two last iterations, probably converged in a single iteration')
            else:
                self.report('STATUS/WARNING: Done, the maximum number of runs '
                            'was reached.\n INFO: The '
                            'charge density of the FLEUR calculation, '
                            f'after {self.ctx.loop_count} FLEUR runs, {num_iterations} '
                            f' iterations and {self.ctx.total_wall_time} sec '
                            f'walltime is {outputnode_dict["distance_charge"]} "me/bohr^3"\n'
                            'INFO: The total energy difference of the last two iterations '
                            f'is {self.ctx.energydiff} Htr and largest force difference is'
                            f'{self.ctx.forcediff} Htr/bohr\n')
        else:  # Termination ok, but not converged yet...
            if self.ctx.abort:  # some error occurred, do not use the output.
                self.report('STATUS/ERROR: I abort, see logs and errors/warning/hints in output_scf_wc_para')

        if self.ctx.last_nmmp_distance > 0.0:
            self.report(f'INFO: The LDA+U density matrix is converged to {self.ctx.last_nmmp_distance} change '
                        'of all matrix elements')

        outputnode_t = Dict(dict=outputnode_dict)
        # this is unsafe so far, because last_calc_out could not exist...
        if last_calc_out:
            outdict = create_scf_result_node(outpara=outputnode_t,
                                             last_calc_out=last_calc_out,
                                             last_calc_retrieved=retrieved)
        else:
            outdict = create_scf_result_node(outpara=outputnode_t)

        # Now it always returns changed fleurinp that was actually used in the calculation
        if self.ctx.fleurinp is not None:
            outdict['fleurinp'] = self.ctx.fleurinp

        if last_calc_out:
            outdict['last_fleur_calc_output'] = last_calc_out

        if self.ctx.last_base_wc:
            self.out_many(self.exposed_outputs(self.ctx.last_base_wc, FleurBaseWorkChain, namespace='last_calc'))

        #outdict['output_scf_wc_para'] = outputnode
        for link_name, node in outdict.items():
            self.out(link_name, node)

        if not self.ctx.reached_conv:
            return self.exit_codes.ERROR_DID_NOT_CONVERGE

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg)  # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_scf_result_node(**kwargs):
    """
    This is a pseudo wf, to create the right graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    for key, val in kwargs.items():
        if key == 'outpara':  # should be always there
            outpara = val
    outdict = {}
    outputnode = outpara.clone()
    outputnode.label = 'output_scf_wc_para'
    outputnode.description = ('Contains self-consistency results and information of an fleur_scf_wc run.')

    outdict['output_scf_wc_para'] = outputnode
    # copy, because we rather produce the same node twice then have a circle in the database for now
    #output_para = args[0]
    # return {'output_eos_wc_para'}
    return outdict



# -*- coding: utf-8 -*-
'fleur_runmax': 4,                   # Maximum number of fleur jobs/starts
'density_converged': 0.00002,        # Charge density convergence criterion
'energy_converged': 0.002,           # Total energy convergence criterion
'force_converged': 0.002,            # Largest force convergence criterion
'mode': 'density',                   # Parameter to converge: 'density', 'force' or 'energy'
'add_comp_para': {
    'only_even_MPI': False,          # True if suppress parallelisation having odd number of MPI
    'max_queue_nodes': 20,           # Max number of nodes allowed (used by automatic error fix)
    'max_queue_wallclock_sec': 86400 # Max number of walltime allowed (used by automatic error fix)
    },
'itmax_per_run': 30,                 # Maximum iterations run for one FleurCalculation
'force_dict': {'qfix': 2,            # parameters required for the 'force' mode
               'forcealpha': 0.5,
               'forcemix': 'BFGS'},
'inpxml_changes': [],                # Modifications to inp.xml



# -*- coding: utf-8 -*-
'beta': {'all': 1.57079},            # see description below
'prop_dir': [1.0, 0.0, 0.0],         # sets a propagation direction of a q-vector
'q_vectors': [[0.0, 0.0, 0.0],       # set a set of q-vectors to calculate SSDispersion
              [0.125, 0.0, 0.0],
              [0.250, 0.0, 0.0],
              [0.375, 0.0, 0.0]],
'ref_qss': [0.0, 0.0, 0.0],          # sets a q-vector for the reference calculation
'inpxml_changes': []                 # additional changes before the FT step
'add_comp_para': {
    'only_even_MPI': False,          # True if suppress parallelisation having odd number of MPI
    'max_queue_nodes': 20,           # Max number of nodes allowed (used by automatic error fix)
    'max_queue_wallclock_sec': 86400 # Max number of walltime allowed (used by automatic error fix)
    },



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Contains helper functions to decide on the paralellization to use for a given system.
'''


def gcd(x, y):
    """
    greatest common divisor
    """
    while y != 0:
        (x, y) = (y, x % y)
    return x


def decide_ncore(nkpoints, ncore):
    """
    For kpoint para the number of cores used must be devidebale
    """
    ncore_new = gcd(nkpoints, ncore)
    ncore_list = list(range(ncore_new, ncore + 1))
    #print ncore_list
    for noc in ncore_list:
        remain = nkpoints % noc
        #print remain
        if not remain:
            ncore_new = noc
    return ncore_new



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This is the worklfow FleurInitialCLSWorkChain 'initial_cls' using the Fleur code calculating
corelevel shifts with different methods.
"""
#TODO parsing of eigenvalues of LOS!
#TODO error handling of scf
#TODO Check if calculations failed, and termine the workflow without a raised execption
# currently the result extraction part will fail if calculations failed
#TODO USE SAME PARAMETERS! (maybe extract method for fleurinp needed)
# TODO: Allow for providing referenes as scf_ouputparameter nodes
# TODO: maybe launch all scfs at the same time
# TODO: gives only a warning currently if ref not found.
# but should lead to error if no ref is found for what should be calculated
from string import digits
from aiida.engine import submit
from aiida.engine import ToContext, WorkChain, if_
from aiida.engine import calcfunction as cf
from aiida.plugins import DataFactory, CalculationFactory
from aiida.orm import Code, load_node, Group, CalcJobNode
from aiida.orm import StructureData, Dict, RemoteData
from aiida.orm import load_group
from aiida.orm.querybuilder import QueryBuilder
from aiida.common.exceptions import NotExistent, MultipleObjectsError
from aiida_fleur.calculation.fleur import FleurCalculation as FleurCalc
from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.tools.common_fleur_wf_util import get_natoms_element
from aiida_fleur.data.fleurinp import FleurinpData


class FleurInitialCLSWorkChain(WorkChain):
    """
    Turn key solution for the calculation of core level shift
    """
    # This block of commented code was removed from the docstring and should be put
    # to the other place in the documentation.
    #     'method' : ['initial', 'full_valence ch', 'half_valence_ch', 'ch', ...]
    #     'Bes' : [W4f, Be1s]
    #     'CLS' : [W4f, Be1s]
    #  toms' : ['all', 'postions' : []]
    #     #'references' : ['calculate', and use # calculate : 'all' , or 'calculate' : ['W', 'Be']
    #     'references' : { 'W': [calc/ouputnode or  fleurinp, or structure data or
    #                      structure data + Parameter  ], 'Be' : }
    #     'scf_para' : {...}, 'default'
    #     'relax' : True
    #     'relax_mode': ['Fleur', 'QE Fleur', 'QE']
    #     'relax_para' : {...}, 'default'
    #     'calculate_doses' : False
    #     'dos_para' : {...}, 'default'

    #     # defaults
    #     default wf_Parameters::
    #     'method' : 'initial'
    #     'atoms' : 'all
    #     'references' : 'calculate'
    #     'scf_para' : 'default'
    #     'relax' : True
    #     'relax_mode': 'QE Fleur'
    #     'relax_para' : 'default'
    #     'calculate_doses' : False
    #     'dos_para' : 'default'
    _workflowversion = '0.5.1'
    _default_wf_para = {
        'references': {},
        'relax': True,
        'relax_mode': 'Fleur',
        'relax_para': 'default',
        'scf_para': 'default',
        'same_para': True,
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        }
    }

    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 6 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        #'max_memory_kb' : None,
        'import_sys_environment': False,
        'environment_variables': {}
    }

    ERROR_INVALID_INPUT_RESOURCES = 1
    ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED = 2
    ERROR_INVALID_CODE_PROVIDED = 3
    ERROR_INPGEN_CALCULATION_FAILED = 4
    ERROR_CHANGING_FLEURINPUT_FAILED = 5
    ERROR_CALCULATION_INVALID_INPUT_FILE = 6
    ERROR_FLEUR_CALCULATION_FALIED = 7
    ERROR_CONVERGENCE_NOT_ARCHIVED = 8
    ERROR_REFERENCE_MISSING = 9

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('wf_parameters', valid_type=Dict, required=False, default=lambda: Dict(dict=cls._default_wf_para))
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.input('inpgen', valid_type=Code, required=False)
        spec.input('structure', valid_type=StructureData, required=False)
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.input('options', valid_type=Dict, required=False)  #, default=Dict(dict=cls._default_options#)
        #)

        spec.outline(cls.check_input, cls.get_references, cls.run_fleur_scfs,
                     if_(cls.relaxation_needed)(cls.relax), cls.find_parameters, cls.run_scfs_ref, cls.return_results)
        spec.output('output_initial_cls_wc_para', valid_type=Dict)

    def check_input(self):
        """
        Init same context and check what input is given if it makes sence
        """
        ### input check ### ? or done automaticly, how optional?

        msg = (
            'INFO: Started initial_state_CLS workflow version {} '
            'Workchain node identifiers: '  #{}"
            ''.format(self._workflowversion))  #, ProcessRegistry().current_calc_node))
        self.report(msg)

        # init
        self.ctx.last_calc = None
        self.ctx.eximated_jobs = 0
        self.ctx.run_jobs = 0
        self.ctx.calcs_res = []
        self.ctx.labels = []
        self.ctx.ref_labels = []
        self.ctx.calcs_torun = []
        self.ctx.ref_calcs_torun = []
        self.ctx.ref_calcs_res = []
        self.ctx.struc_to_relax = []
        self.ctx.successful = True
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.ref = {}
        self.ctx.calculate_formation_energy = True

        #Style: {atomtype : listof all corelevel, atomtype_coresetup... }
        #ie: { 'W-1' : [shift_1s, ... shift 7/2 4f],
        #      'W-1_coreconfig' : ['1s','2s',...],
        #      'W-2' : [...], 'Be-1': [], ...} #all in eV!
        self.ctx.CLS = {}
        self.ctx.cl_energies = {}  # same style as CLS only energy <-> shift
        self.ctx.ref_cl_energies = {}
        #Style: {'Compound' : energy, 'ref_x' : energy , ...}
        #i.e {'Be12W' : 0.0, 'Be' : 0.104*htr_eV , 'W' : 0.12*htr_eV} # all in eV!
        self.ctx.fermi_energies = {}
        self.ctx.bandgaps = {}
        self.ctx.atomtypes = {}
        # set values, or defaults for Wf_para
        # 'wf_parameters' always there
        wf_dict = self.inputs.wf_parameters.get_dict()
        default = self._default_wf_para

        self.ctx.add_comp_para = wf_dict.get('add_para_calc', default.get('add_para_calc'))
        self.ctx.same_para = wf_dict.get('same_para', default.get('same_para'))
        self.ctx.scf_para = wf_dict.get('scf_para', default.get('scf_para'))
        self.ctx.relax = wf_dict.get('relax', default.get('relax'))
        self.ctx.relax_mode = wf_dict.get('relax_mode', default.get('relax_mode'))
        self.ctx.relax_para = wf_dict.get('relax_para', default.get('dos_para'))

        defaultoptions = self._default_options
        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions
        for key, val in defaultoptions.items():
            options[key] = options.get(key, val)
        self.ctx.options = options

        # check if inputs given make sense # TODO sort this out in common wc
        inputs = self.inputs
        if 'fleurinp' in inputs:
            #TODO make a check if an extracted structure exists, since get_structuredata is wf
            structure = inputs.fleurinp.get_structuredata(inputs.fleurinp)
            self.ctx.elements = list(structure.get_composition().keys())
            self.ctx.calcs_torun.append(inputs.get('fleurinp'))
            #print('here1')
            if 'structure' in inputs:
                warning = 'WARNING: Ignoring Structure input, because Fleurinp was given'
                self.ctx.warnings.append(warning)
                self.report(warning)
            if 'calc_parameters' in inputs:
                warning = 'WARNING: Ignoring parameter input, because Fleurinp was given'
                self.ctx.warnings.append(warning)
                self.report(warning)
        elif 'structure' in inputs:
            self.ctx.elements = list(inputs.structure.get_composition().keys())
            #self.ctx.elements = list(s.get_symbols_set())
            if 'inpgen' not in inputs:
                error = 'ERROR: StructureData was provided, but no inpgen code was provided'
                self.ctx.errors.append(error)
                self.control_end_wc(error)
                return self.ERROR_INVALID_INPUT_RESOURCES
            if 'calc_parameters' in inputs:
                self.ctx.calcs_torun.append([inputs.get('structure'), inputs.get('calc_parameters')])
            else:
                self.ctx.calcs_torun.append(inputs.get('structure'))
        else:
            error = 'ERROR: No StructureData nor FleurinpData was provided'
            self.ctx.errors.append(error)
            self.control_end_wc(error)
            return self.ERROR_INVALID_INPUT_RESOURCES
        self.report(f'INFO: elements in structure: {self.ctx.elements}')

    def get_references(self):
        """
        To calculate a CLS in initial state approx, we need reference calculations
        to the Elemental crystals. First it is checked if the user has provided them
        Second the database is checked, if there are structures with certain extras.
        Third the COD database is searched for the elemental Cystal structures.
        If some referneces are not found stop here.
        Are there already calculation of these 'references', ggf use them.
        We do not put these calculation in the calculation queue yet because we
        need specific parameters for them
        """

        self.report('INFO: In Get_references initial_state_CLS workflow')

        references = self.inputs.wf_parameters.get_dict().get('references', {})
        # should be of the form of
        #'references' : { 'W': calc, outputnode of workflow or fleurinp,
        #                 or structure data or (structure data + Parameter),
        #                 'Be' : ...}
        self.ctx.ref_calcs_torun = []
        self.ctx.ref = {}
        self.ctx.abort = False

        struc_group = references.get('group', None)
        para_group = references.get('para_group', None)

        #TODO better checks if ref makes sense?

        # get specific element reference if given override
        #print(self.ctx.elements)
        elements = self.ctx.elements  # ggf copy because ctx.elements will be modified
        for elem in elements:
            #to_calc[elem] = 'find'
            ref_el = references.get(elem, None)
            #print ref_el
            if ref_el:
                # loading nodes
                if isinstance(ref_el, list):
                    ref_el_node = []
                    for ref_el_el in ref_el:
                        try:
                            ref_el_nodes = load_node(ref_el_el)
                        except (NotExistent, MultipleObjectsError, ValueError, TypeError):
                            ref_el_node = None
                            self.report('ERROR: The reference node in the list '
                                        '(id or uuid) provided: {} for element: '
                                        '{} could not be loaded with load_node'
                                        ''.format(ref_el_el, elem))
                            self.ctx.abort = True
                        ref_el_node.append(ref_el_nodes)
                else:
                    try:
                        ref_el_node = load_node(ref_el)
                    except (NotExistent, MultipleObjectsError, ValueError, TypeError):
                        # NotExistent: No node was found
                        ref_el_node = None
                        self.report('ERROR: The reference node (id or uuid) '
                                    'provided: {} for element: {} could'
                                    'not be loaded with load_node'
                                    ''.format(ref_el, elem))
                        self.ctx.abort = True

                # expecting nodes and filling ref_calcs_torun
                if isinstance(ref_el_node, list):  #(StructureData, Dict)):
                    #enforced parameters, add directly to run queue
                    # TODO: if a scf with these parameters was already done link to it
                    # and extract the results instead of running the calculation again....
                    if len(ref_el_node) == 2:
                        if isinstance(ref_el_node[0], StructureData) and isinstance(ref_el_node[1], Dict):
                            self.ctx.ref_calcs_torun.append(ref_el_node)
                        else:
                            self.report('WARNING: I did not undestand the list with length 2 '
                                        'you gave me as reference input')
                    else:
                        self.report('WARNING: I did not undestand the list {} with length {} '
                                    'you gave me as reference input'
                                    ''.format(ref_el_node, len(ref_el_node)))
                elif isinstance(ref_el_node, CalcJobNode):
                    #extract from fleur calc TODO
                    self.ctx.ref_cl_energies[elem] = {}
                elif isinstance(ref_el_node, Dict):
                    #extract from workflow output TODO
                    self.ctx.ref_cl_energies[elem] = {}
                elif isinstance(ref_el_node, FleurinpData):
                    # add to calculations
                    #enforced parameters, add directly to run queue
                    self.ctx.ref_calcs_torun.append(ref_el_node)
                    #self.ctx.ref[elem] = ref_el
                elif isinstance(ref_el_node, StructureData):
                    self.ctx.ref[elem] = ref_el_node
                    self.ctx.ref_calcs_torun.append(ref_el_node)
                #elif isinstance(ref_el, initial_state_CLS):
                #    extract TODO
                else:
                    error = f'ERROR: I do not know what to do with this given reference {ref_el} for element {elem}'
                    #print(error)
                    self.report(error)
                    self.ctx.errors.append(error)
                    self.ctx.abort = True
            elif struc_group:
                #print('here, looking in group')
                #print(elem, struc_group)
                structure, report = get_ref_from_group(elem, struc_group)
                if report:
                    self.report(report)
                parameter, report = get_para_from_group(elem, para_group)
                if structure and parameter:
                    self.ctx.ref[elem] = structure
                    self.ctx.ref_calcs_torun.append([structure, parameter])
                elif structure:
                    self.ctx.ref[elem] = structure
                    self.ctx.ref_calcs_torun.append(structure)
                else:
                    pass  # report not found?

            #elif query_for_ref: # no ref given, we have to look for it.
            #    structure = querry_for_ref_structure(elem)
            #    #TODO: Advance this querry, if a calculation with the given
            #    #parameter was already done use these results
            #    if structure:
            #        self.ctx.ref[elem] = structure
            #        self.ctx.ref_calcs_torun.append(structure)# tempoary later check parameters
            #    else: #not found
            #        error = ("ERROR: Reference structure for element: {} not found."
            #                 "checkout the 'querry_for_ref_structure' method."
            #                 "to see what extras are querried for.".format(elem))
            #        #print(error)
            #        self.ctx.errors.append(error)
            #        self.ctx.abort = True
            #        self.report(error)

            else:  # no reference for element found
                # do we not want to calculate it or is this an error?
                warning = ('WARNING: I did not find a reference for element {}. '
                           'If you do not calculate shifts for this element '
                           'ignore this warning. If I should calculate this will '
                           'lead to an error later. Note: without all references'
                           ' I cannot calculte the binding energy'.format(elem))
                self.report(warning)
                self.ctx.calculate_formation_energy = False
                # delete element from element list (no calculations will be launched for it)
                valid_elm = self.ctx.elements
                i = valid_elm.index(elem)
                del valid_elm[i]
                self.ctx.elements = valid_elm

        if self.ctx.abort:
            error = ('ERROR: Something was wrong with the reference input provided. '
                     'I cannot calculate from the input, or what I have found '
                     'what you want me to do. Please check the workchain report'
                     'for details.')
            self.control_end_wc(error)
            return self.ERROR_REFERENCE_MISSING

        self.report(f'INFO ref_calcs_torun: {self.ctx.ref_calcs_torun} ')

        # check if a structureData for these elements was given
        #if yes add to ref_calc to run
        #was also a prameter node given for the element?
        #yes run with these
        #no was on given for the host structure, extract element parameternode

        #else use parameters extracted from host calculation # TODO

        #check if there is a structure from this element in the database with extras:
        # with extra.type = 'bulk', extra.specific = 'reference',
        #'extra.elemental' = True, extra.structure = 'W'
        # check if input parameter node values for the calculation are the same.

        #if yes, if a calculation exists use that result
        #else do a calculation on that structure as above

    def run_fleur_scfs(self):
        """
        Run SCF-cycles for all structures, calculations given in certain workflow arrays.
        """
        self.report('INFO: In run_fleur_scfs initial_state_CLS workflow')

        para = self.ctx.scf_para
        if para == 'default':
            wf_parameter = {}
        else:
            wf_parameter = para
        wf_parameter['add_comp_para'] = self.ctx.add_comp_para
        #wf_parameter['options'] = self.ctx.options
        wf_parameters = Dict(dict=wf_parameter)
        resall = {}
        calc_labels = []
        options = Dict(dict=self.ctx.options)
        # for each calulation in self.ctx.calcs_torun #TODO what about wf params?
        res = None
        #print(self.ctx.calcs_torun)
        for i, node in enumerate(self.ctx.calcs_torun):
            #print node
            calclabel = f'ori_calc_{i}'
            calc_labels.append(calclabel)
            scf_label = 'cls|scf_wc main'
            scf_description = 'cls|scf of the main structure'
            if isinstance(node, StructureData):
                inputs = {
                    'wf_parameters': wf_parameters,
                    'structure': node,
                    'options': options,
                    'inpgen': self.inputs.inpgen,
                    'fleur': self.inputs.fleur,
                    #'metadata' : {'label' : scf_label,
                    #             'description' : scf_description}
                }
                res = self.submit(FleurScfWorkChain, **inputs)
            #elif isinstance(node, FleurinpData):
            #    res = fleur_scf_wc.run(wf_parameters=wf_parameters, structure=node,
            #                inpgen = self.inputs.inpgen, fleur=self.inputs.fleur)#
            elif isinstance(node, list):  #(StructureData, Dict)):
                if len(node) == 2:
                    inputs = {
                        'wf_parameters': wf_parameters,
                        'structure': node[0],
                        'calc_parameters': node[1],
                        'options': options,
                        'inpgen': self.inputs.inpgen,
                        'fleur': self.inputs.fleur,
                        'metadata': {
                            'label': scf_label,
                            'description': scf_description
                        }
                    }
                    res = self.submit(FleurScfWorkChain, **inputs)
                else:
                    self.report('ERROR: something in calcs_torun which I do not'
                                'recognize, list has not 2 entries: {}'.format(node))
            else:
                self.report(f'ERROR: something in calcs_torun which I do not recognize: {node}')
                #self.report('{}{}'.format(type(node[0], node[1])))
                res = None
                continue
            resall[calclabel] = res  #append(res)
            #print res
            #calc_node = res['output_scf_wc_para'].get_inputs()[0]
            # if run is used, otherwise use labels
            #self.ctx.calcs_res.append(calc_node)
            #self.ctx.calcs_torun.remove(node)
            #print res
        self.ctx.calcs_torun = []
        self.ctx.calc_labels = calc_labels

        print('res_all', resall)
        #return ToContext(last_calc=res)
        '''
        inputs = get_inputs_fleur(code, remote, fleurin, options)
        future = submit(FleurCalc, **inputs)
        self.ctx.loop_count = self.ctx.loop_count + 1
        print 'run FLEUR number: {}'.format(self.ctx.loop_count)
        self.ctx.calcs.append(future)
        '''
        return ToContext(**resall)  #calcs_res=res)  #calcs.append(future

    def relaxation_needed(self):
        """
        If the structures should be relaxed, check if their Forces are below a certain
        threshold, otherwise throw them in the relaxation wf.
        """
        self.report('INFO: In relaxation initial_state_CLS workflow (so far nothing to do)')
        if self.ctx.relax:
            # TODO check all forces of calculations
            forces_fine = True
            return forces_fine
        else:
            return False

    def relax(self):
        """
        Do structural relaxation for certain structures.
        """
        self.report('INFO: In relax initial_state_CLS workflow (so far nothing to do)')
        self.ctx.dos_to_calc = []
        for calc in self.ctx.dos_to_calc:
            pass
            # TODO run relax workflow

    def find_parameters(self):
        """
        If the same parameters shall be used in the calculations you have to
        find some that match. For low error on CLS. therefore use the ones enforced
        or extract from the previous Fleur calculation.
        """
        #self.ctx.ref[elem] = ref_el
        #self.ctx.ref_calcs_torun.append(ref_el)

        # for entry in ref[elem] find parameter node
        for elm, struc in self.ctx.ref.items():
            #print(elm, struc)
            #self.ctx.ref_calcs_torun.append(ref_el)
            pass
            # if parameter node given, extract from there,
            #parameter_dict
            # else
            #extract parameter out of previous calculation
            #parameter_dict = fleurinp.extract_para(element)
            # BE CAREFUL WITH LOs! soc and co

    def run_scfs_ref(self):
        """
        Run SCF-cycles for ref structures, calculations given in certain workflow arrays.
        parameter nodes should be given
        """
        self.report('INFO: In run_scfs_ref initial_state_CLS workflow')

        para = self.ctx.scf_para
        if para == 'default':
            wf_parameter = {}
        else:
            wf_parameter = para
        wf_parameter['add_comp_para'] = self.ctx.add_comp_para
        # TODO maybe use less resources, or default of one machine
        #wf_parameter['options'] = self.ctx.options
        wf_parameters = Dict(dict=wf_parameter)
        options = Dict(dict=self.ctx.options)

        res_all = []
        calcs = {}
        # now in parallel
        #print self.ctx.ref_calcs_torun
        i = 0
        #print(self.ctx.ref_calcs_torun)
        for i, node in enumerate(self.ctx.ref_calcs_torun):
            scf_label = f'cls|scf_wc on ref {self.ctx.elements[i]}'
            scf_description = f'cls|scf of the reference structure of element {self.ctx.elements[i]}'
            #print node
            if isinstance(node, StructureData):
                inputs = {
                    'wf_parameters': wf_parameters,
                    'structure': node,
                    'options': options,
                    'inpgen': self.inputs.inpgen,
                    'fleur': self.inputs.fleur,
                    'metadata': {
                        'label': scf_label,
                        'description': scf_description
                    }
                }
                res = self.submit(FleurScfWorkChain, **inputs)  #

            #elif isinstance(node, FleurinpData):
            #    res = submit(fleur_scf_wc, wf_parameters=wf_parameters, structure=node,
            #                inpgen = self.inputs.inpgen, fleur=self.inputs.fleur)#
            elif isinstance(node, list):  #(StructureData, Dict)):
                inputs = {
                    'wf_parameters': wf_parameters,
                    'structure': node[0],
                    'calc_parameters': node[1],
                    'options': options,
                    'inpgen': self.inputs.inpgen,
                    'fleur': self.inputs.fleur,
                    'metadata': {
                        'label': scf_label,
                        'description': scf_description
                    }
                }
                res = self.submit(FleurScfWorkChain, **inputs)  #
            else:
                self.report(f'WARNING: something in calcs_torun which I do not reconise: {node}')
                continue
            label = str(f'calc_ref{i}')
            #print(label)
            #calc_node = res['output_scf_wc_para'].get_inputs()[0] # if run is used, otherwise use labels
            self.ctx.ref_labels.append(label)
            calcs[label] = res
            res_all.append(res)
            self.ctx.ref_calcs_res.append(res)
            #self.ctx.calcs_torun.remove(node)
            #print res
        self.ctx.ref_calcs_torun = []
        return ToContext(**calcs)

    def handle_scf_failure(self):
        """
        In here we handle all failures from the scf workchain
        """
        '''
        try:
            calculation = self.ctx.calculation
        except Exception:
            self.abort_nowait('the first iteration finished without returning a PwCalculation')
            return

        expected_states = [calc_states.FINISHED, calc_states.FAILED, calc_states.SUBMISSIONFAILED]

        # Done: successful convergence of last calculation
        if calculation.has_finished_ok():
            self.report('converged successfully after {} iterations'.format(self.ctx.iteration))
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True

        # Abort: exceeded maximum number of retries
        elif self.ctx.iteration >= self.ctx.max_iterations:
            self.report('reached the max number of iterations {}'.format(self.ctx.max_iterations))
            self.abort_nowait('last ran PwCalculation<{}>'.format(calculation.pk))

        # Abort: unexpected state of last calculation
        elif calculation.get_state() not in expected_states:
            self.abort_nowait('unexpected state ({}) of PwCalculation<{}>'.format(
                calculation.get_state(), calculation.pk))

        # Retry: submission failed, try to restart or abort
        elif calculation.get_state() in [calc_states.SUBMISSIONFAILED]:
            self._handle_submission_failure(calculation)

        # Retry: calculation failed, try to salvage or abort
        elif calculation.get_state() in [calc_states.FAILED]:
            self._handle_calculation_failure(calculation)

        # Retry: try to convergence restarting from this calculation
        else:
            self.report('calculation did not converge after {} iterations, restarting'.format(self.ctx.iteration))
            self.ctx.restart_calc = calculation

        '''

        failure = False
        if failure:
            self.ctx.successful = False

    def collect_results(self):
        """
        Collect results from certain calculation, check if everything is fine,
        calculate the wanted quantities. currently all energies are in hartree
        (as provided by Fleur)
        """
        from aiida_fleur.tools.common_fleur_wf_util import determine_formation_energy

        message = ('INFO: Collecting results of initial_state_CLS workflow')
        self.report(message)
        # TODO be very careful with core config?
        #from pprint import pprint

        #self.ctx.ref_cl_energies
        all_CLS = {}
        # get results from calc
        calcs = []  #self.ctx.calcs_res
        ref_calcs = []  #self.ctx.ref_calcs_res
        for label in self.ctx.ref_labels:
            calc = self.ctx[label]
            ref_calcs.append(calc)
        for label in self.ctx.calc_labels:
            calc = self.ctx[label]
            calcs.append(calc)
        # extract_results need the scf workchain calculation node
        total_en, fermi_energies, bandgaps, atomtypes, all_corelevel, log_ref = extract_results(calcs)
        ref_total_en, ref_fermi_energies, ref_bandgaps, ref_atomtypes, ref_all_corelevel, log = extract_results(
            ref_calcs)

        if log_ref or log:
            self.ctx.successful = False
            self.ctx.warnings.append(log_ref)
            self.ctx.warnings.append(log)
        #print(all_corelevel)
        #print(ref_all_corelevel)

        ref_cl_energies = {}
        cl_energies = {}

        #first substract efermi from corelevel of reference structures
        # TODO check if both values, corelevel and efermi are in eV
        for compound, atomtypes_list in ref_atomtypes.items():
            # atomtype_list contains a list of dicts of all atomtypes from compound x
            # get corelevels of compound x
            cls_all_atomtyps = ref_all_corelevel[compound]
            for i, atomtype in enumerate(atomtypes_list):
                #atomtype a dict which contains one atomtype
                elm = atomtype.get('element', None)
                cls_atomtype = cls_all_atomtyps[i][0]
                ref_cl_energies[elm] = []
                ref_cls = []
                for corelevel in cls_atomtype['corestates']:
                    ref_cls.append(corelevel['energy'] - ref_fermi_energies[compound])
                ref_cl_energies[elm].append(ref_cls)

        #print('ref_cl energies')
        #print(ref_cl_energies)
        #print(all_corelevel)

        #now substract efermi from corelevel of compound structure
        #and calculate core level shifts
        for compound, cls_atomtypes_list in all_corelevel.items():
            #init, otherwise other types will override
            for i, atomtype in enumerate(atomtypes[compound]):
                elm = atomtype.get('element', None)
                cl_energies[elm] = []
                all_CLS[elm] = []

            #now fill
            for i, atomtype in enumerate(atomtypes[compound]):
                elm = atomtype.get('element', None)
                #print elm
                cls_atomtype = cls_atomtypes_list[i]
                corelevels = []
                for corelevel in cls_atomtype[0]['corestates']:
                    correct_cl = corelevel['energy'] - fermi_energies[compound]
                    corelevels.append(correct_cl)
                cl_energies[elm].append(corelevels)

                #now calculate CLS
                ref = ref_cl_energies.get(elm, [0])[-1]  # We just use one (last) atomtype
                #of elemental reference (in general might be more complex,
                #since certain elemental unit cells could have several atom types (graphene))
                corelevel_shifts = []
                #TODO shall we store just one core-level shift per atomtype?
                for j, corelevel in enumerate(cl_energies[elm][-1]):
                    corelevel_shifts.append(corelevel - float(ref[j]))
                all_CLS[elm].append(corelevel_shifts)

        # calculate formation energy
        #determine_formation_energy({'BeW' : 2, 'Be2W' : 2.5}, {'Be' : 1, 'W' : 1})
        # to normalize ref,
        # from.split(012345678910)
        # devide total energy by number of atoms

        # Formation energy calculation is ony possible if all elementals of the structure
        # have been calculated.
        # convert total_en dict to list, why?
        total_en_list = []
        for key, val in total_en.items():
            total_en_list.append([key, val])
        if self.ctx.calculate_formation_energy:
            # the reference total energy is for the whole structure with several atoms,
            # we need it per atom
            ref_total_en_norm = {}
            for key, val in ref_total_en.items():
                elm_dict = get_natoms_element(key)
                ref_total_en_norm[list(elm_dict.keys())[0]] = 1.0 * val / list(elm_dict.values())[0]
            #print(ref_total_en_norm)
            #print(total_en)

            formation_energy, form_dict = determine_formation_energy(total_en, ref_total_en_norm)
            #print(form_dict)
        else:
            formation_energy = [[]]

        # TODO make simpler format of atomtypes for node
        # TODO write corelevel explanation/coresetup in a format like 4f7/2
        #TODO ? also get total energies?
        return cl_energies, all_CLS, ref_cl_energies, fermi_energies, bandgaps, ref_fermi_energies, ref_bandgaps, atomtypes, ref_atomtypes, formation_energy, total_en, ref_total_en
        #return cl_energies, all_CLS, ref_cl_energies, fermi_energies, bandgaps, ref_fermi_energies, ref_bandgaps, atomtypes, ref_atomtypes, formation_energy[0], total_en, ref_total_en

    def return_results(self):
        """
        return the results of the calculations
        """
        # TODO more output, info here

        #print corelevel shifts were calculated bla bla
        cl, cls, ref_cl, efermi, gap, ref_efermi, ref_gap, at, at_ref, formE, tE, tE_ref = self.collect_results()

        if self.ctx.errors:
            self.ctx.warnings.append(self.ctx.errors)

        material = list(efermi.keys())
        if material:
            material = material[0]
        fermi_energy = list(efermi.values())
        if fermi_energy:
            fermi_energy = fermi_energy[0]
        total_energy = list(tE.values())
        if total_energy:
            total_energy = total_energy[0]
        bandgap = list(gap.values())
        if bandgap:
            bandgap = bandgap[0]

        outputnode_dict = {}

        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['material'] = material
        outputnode_dict['corelevel_energies'] = cl
        outputnode_dict['corelevel_energies_units'] = 'htr'  #'eV'
        outputnode_dict['reference_corelevel_energies'] = ref_cl
        outputnode_dict['reference_corelevel_energies_units'] = 'htr'  #'eV'
        outputnode_dict['reference_fermi_energy'] = list(ref_efermi.values())
        outputnode_dict['reference_fermi_energy_des'] = list(ref_efermi.keys())
        outputnode_dict['fermi_energy'] = efermi
        outputnode_dict['fermi_energy_units'] = 'htr'
        outputnode_dict['corelevelshifts'] = cls
        outputnode_dict['corelevelshifts_units'] = 'htr'
        outputnode_dict['binding_energy_convention'] = 'negativ'
        #outputnode_dict['coresetup'] = []#cls
        #outputnode_dict['reference_coresetup'] = []#cls
        outputnode_dict['bandgap'] = bandgap
        outputnode_dict['bandgap_units'] = 'htr'
        outputnode_dict['reference_bandgaps'] = list(ref_gap.values())
        outputnode_dict['reference_bandgaps_des'] = list(ref_gap.keys())
        outputnode_dict['atomtypes'] = at
        outputnode_dict['formation_energy'] = formE
        outputnode_dict['formation_energy_units'] = 'eV/atom'
        outputnode_dict['total_energy'] = total_energy
        outputnode_dict['total_energy_units'] = 'eV'
        outputnode_dict['total_energy_ref'] = list(tE_ref.values())
        outputnode_dict['total_energy_ref_des'] = list(tE_ref.keys())
        #outputnode = Dict(dict=outputnode_dict)

        # To have to ouput node linked to the calculation output nodes
        outnodedict = {}
        outnode = Dict(dict=outputnode_dict)
        outnodedict['results_node'] = outnode

        # TODO: bad design, put in calcfunction and make bullet proof.
        calc = self.ctx[self.ctx.calc_labels[-1]]
        calc_dict = calc.get_outgoing().get_node_by_label('output_scf_wc_para')
        outnodedict['input_structure'] = calc_dict

        for label in self.ctx.ref_labels:
            calc = self.ctx[label]
            calc_dict = calc.get_outgoing().get_node_by_label('output_scf_wc_para')
            outnodedict[label] = calc_dict

        outdict = create_initcls_result_node(**outnodedict)

        #outdict = {}
        #outdict['output_initial_cls_wc_para'] = outputnode
        #print outdict
        for k, v in outdict.items():
            self.out(k, v)
        msg = ('INFO: Initial_state_CLS workflow Done')
        self.report(msg)

    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain. will initalize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg)  # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def create_initcls_result_node(**kwargs):
    """
    This is a pseudo wf, to create the rigth graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """

    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_initial_cls_wc_para'] = outpara.clone()
    # copy, because we rather produce the same node twice then have a circle in the database for now...
    #output_para = args[0]
    #return {'output_eos_wc_para'}
    return outdict


def querry_for_ref_structure(element_string):
    """
    This methods finds StructureData nodes with the following extras:
    extra.type = 'bulk', # Should be done by looking at pbc, but I could not
    get querry to work.
    extra.specific = 'reference',
    'extra.elemental' = True,
    extra.structure = element_string

    param: element_string: string of an element
    return: the latest StructureData node that was found
    """

    #query db
    q = QueryBuilder()
    q.append(StructureData,
             filters={
                 'extras.type': {
                     '==': 'bulk'
                 },
                 'extras.specification': {
                     '==': 'reference'
                 },
                 'extras.elemental': {
                     '==': True
                 },
                 'extras.element': {
                     '==': element_string
                 }
             })
    q.order_by({StructureData: 'ctime'})  #always use the most recent
    structures = q.all()

    if structures:
        return structures[-1][0]
    else:
        return None


def fleur_calc_get_structure(calc_node):
    """
    Get the AiiDA data structure from a fleur calculations
    """
    #get fleurinp
    fleurinp = calc_node.inp.fleurinpdata
    structure = fleurinp.get_structuredata(fleurinp)
    return structure


def extract_results(calcs):
    """
    Collect results from certain calculation, check if everything is fine,
    calculate the wanted quantities.

    params: calcs : list of scf workchains nodes
    """

    from aiida_fleur.tools.extract_corelevels import extract_corelevels
    log = []
    calc_uuids = []
    for calc in calcs:
        #print(calc)
        try:
            calc_uuid = calc.get_outgoing().get_node_by_label('output_scf_wc_para').get_dict()['last_calc_uuid']
        except (NotExistent, MultipleObjectsError, ValueError, TypeError, KeyError):  #TODO which error
            logmsg = ('ERROR: No output_scf_wc_para node found or no "last_calc_uuid" '
                      'key in it for calculation: {}'.format(calc))
            log.append(logmsg)
            continue
        if calc_uuid is not None:
            calc_uuids.append(calc_uuid)
        #calc_uuids.append(calc['output_scf_wc_para'].get_dict()['last_calc_uuid'])

    all_corelevels = {}
    fermi_energies = {}
    bandgaps = {}
    all_atomtypes = {}
    total_energy = {}

    # more structures way: divide into this calc and reference calcs.
    # currently the order in calcs is given, but this might change if you submit
    # check if calculation pks belong to successful fleur calculations
    for uuid in calc_uuids:
        calc = load_node(uuid)
        if not isinstance(calc, CalcJobNode):
            #raise ValueError("Calculation with pk {} must be a FleurCalculation".format(pk))
            # log and continue
            continue
        if calc.is_finished_ok:
            # get out.xml file of calculation
            #outxml = calc.outputs.retrieved.folder.get_abs_path('path/out.xml')
            outxml = calc.outputs.retrieved.open('out.xml')
            #print outxml
            try:
                corelevels, atomtypes = extract_corelevels(outxml)
            finally:
                outxml.close()
            #all_corelevels.append(core)
            #print('corelevels: {}'.format(corelevels))
            #print('atomtypes: {}'.format(atomtypes))
            #for i in range(0,len(corelevels[0][0]['corestates'])):
            #    print corelevels[0][0]['corestates'][i]['energy']

            #TODO how to store?
            efermi = calc.res.fermi_energy
            #print efermi
            bandgap = calc.res.bandgap
            te = calc.res.energy
            #total_energy = calc.res.total_energy
            #total_energy_units = calc.res.total_energy_units

            # TODO: maybe different, because it is prob know from before
            #fleurinp = calc.inp.fleurinpdata
            #structure = fleurinp.get_structuredata(fleurinp)
            #compound = structure.get_formula()
            #print compound
            #number = '{}'.format(i)
            #fermi_energies[number] = efermi
            #bandgaps[number] = bandgap
            #all_atomtypes[number] = atomtypes
            #all_corelevels[number] = corelevels
            #all_total_energies[number] = total_energy
        else:
            # log and continue
            te = float('nan')
            bandgap = float('nan')
            efermi = float('nan')
            corelevels = [float('nan')]
            atomtypes = [float('nan')]
            logmsg = f'ERROR: Fleur Calculation with uuid {uuid} was not in in state FINISHED'
            log.append(logmsg)
            #continue
            #raise ValueError("Calculation with pk {} must be in state FINISHED".format(pk))

        # TODO: maybe different, because it is prob know from before
        fleurinp = calc.inputs.fleurinpdata
        structure = fleurinp.get_structuredata_ncf()
        compound = structure.get_formula()
        #print compound
        fermi_energies[compound] = efermi
        bandgaps[compound] = bandgap
        all_atomtypes[compound] = atomtypes
        all_corelevels[compound] = corelevels
        total_energy[compound] = te
        #fermi_energies = efermi
        #bandgaps = bandgap
        #all_atomtypes = atomtypes
        #all_corelevels = corelevels

    return total_energy, fermi_energies, bandgaps, all_atomtypes, all_corelevels, log
    #TODO validate results and give some warnings

    # check bandgaps, if not all metals, throw warnings:
    # bandgap and efermi prob wrong, which makes some results meaningless

    # check fermi energy differences, correct results for fermi energy diff
    # ggf TODO make a raw core-level and core-level to fermi energy variable
    #TODO to what reference energy? or better not to fermi, but first unocc? (add bandgap)

    #Style: {atomtype : listof all corelevel, atomtype_coresetup... }
    #ie: { 'W-1' : [shift_1s, ... shift 7/2 4f],
    #      'W-1_coreconfig' : ['1s','2s',...],
    #      'W-2' : [...], 'Be-1': [], ...} #all in eV!
    #self.ctx.CLS = {}
    #self.ctx.cl_energies = {}# same style as CLS only energy <-> shift

    #Style: {'Compound' : energy, 'ref_x' : energy , ...}
    #i.e {'Be12W' : 0.0, 'Be' : 0.104*htr_eV , 'W' : 0.12*htr_eV} # all in eV!
    #self.ctx.fermi_energies = {}


def get_ref_from_group(element, group):
    """
    Return a structure data node from a given group for a given element.
    (quit creedy, done straighforward)

    params: group: group name or pk
    params: element: string with the element i.e 'Si'

    returns: AiiDA StructureData node
    """

    report = []
    group_name = None
    try:
        group_pk = int(group)
    except ValueError:
        group_pk = None
        group_name = group

    if group_pk is not None:
        try:
            str_group = Group(dbgroup=group_pk)
        except NotExistent:
            str_group = None
            message = ('You have to provide a valid pk for a Group of'
                       'structures or a Group name. Reference key: "group".'
                       'given pk= {} is not a valid group'
                       '(or is your group name integer?)'.format(group_pk))
            report.append(message)
    else:
        try:
            str_group = Group.get_from_string(group_name)
        except NotExistent:
            str_group = None
            message = ('You have to provide a valid pk for a Group of'
                       'structures or a Group name. Wf_para key: "struc_group".'
                       'given group name= {} is not a valid group'
                       '(or is your group name integer?)'.format(group_name))
            report.append(message)
            #abort_nowait('I abort, because I have no structures to calculate ...')

    stru_nodes = str_group.nodes
    #n_stru = len(stru_nodes)

    structure = None

    for struc in stru_nodes:
        formula = struc.get_formula()
        eformula = formula.translate(None, digits)  # remove digits, !python3 differs
        if eformula == element:
            return struc, report

    report.append(f'Structure node for element {element} not found in group {group}')

    return structure, report


def get_para_from_group(element, group):
    """
    get structure node for a given element from a given group of structures
    (quit creedy, done straighforward)
    """

    report = []

    group_name = None
    try:
        group_pk = int(group)
    except ValueError:
        group_pk = None
        group_name = group

    if group_pk is not None:
        try:
            para_group = load_group(pk=group_pk)
        except NotExistent:
            para_group = None
            message = ('You have to provide a valid pk for a Group of '
                       'parameters or a Group name. Reference key: "group".'
                       'given pk= {} is not a valid group'
                       '(or is your group name integer?)'.format(group_pk))
            report.append(message)
    else:
        try:
            para_group = load_group(label=group_name)
        except NotExistent:
            para_group = None
            message = ('You have to provide a valid label for a Group of '
                       'parameters or a Group name. Wf_para key: "para_group".'
                       'given group name= {} is not a valid group'
                       '(or is your group name integer?)'.format(group_name))
            report.append(message)
            #abort_nowait('I abort, because I have no structures to calculate ...')

    para_nodes = para_group.nodes
    #n_stru = len(para_nodes)

    parameter = None

    for para in para_nodes:
        formula = para.get_extras().get('element', None)
        #eformula = formula.translate(None, digits) # remove digits, !python3 differs
        if formula == element:
            return para, report

    report.append(f'Parameter node for element {element} not found in group {group}')

    return parameter, report


def clshifts_to_be(coreleveldict, reference_dict):
    """
    This methods converts corelevel shifts to binding energies,
    if a reference is given.
    These can than be used for plotting.

    Example:

    .. code-block:: python

            reference = {'W' : {'4f7/2' : [124],
                               '4f5/2' : [102]},
                         'Be' : {'1s': [117]}}
            corelevels = {'W' : {'4f7/2' : [0.4, 0.3, 0.4 ,0.1],
                                 '4f5/2' : [0, 0.3, 0.4, 0.1]},
                          'Be' : {'1s': [0, 0.2, 0.4, 0.1, 0.3]}}

    """

    return_corelevel_dict = {}
    for elem, corelevel_dict in coreleveldict.items():
        ref_el = reference_dict.get(elem, {})
        return_corelevel_dict[elem] = {}
        for corelevel_name, corelevel_list in corelevel_dict.items():
            ref_cl = ref_el.get(corelevel_name, [])
            be_all = []
            nref = len(ref_cl)
            ncl = len(corelevel_list)
            if nref == ncl:
                for i, corelevel in enumerate(corelevel_list):
                    be = corelevel + ref_cl[i]
                    be_all.append(be)
            else:
                for corelevel in corelevel_list:
                    be = corelevel + ref_cl[0]
                    be_all.append(be)
            return_corelevel_dict[elem][corelevel_name] = be_all

    return return_corelevel_dict



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This is the worklfow 'dos' for the Fleur code, which calculates a
density of states (DOS).
"""
import os.path

from aiida.plugins import DataFactory
from aiida.orm import Code, StructureData, Dict, RemoteData
from aiida.engine import WorkChain, ToContext
from aiida.engine import submit
#from aiida.work.process_registry import ProcessRegistry
from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.data.fleurinp import FleurinpData


class fleur_dos_wc(WorkChain):
    """
    DEPRECATED: Use FleurBandDosWorkChain instead (entrypoint fleur.banddos)
    This workflow calculated a DOS from a Fleur calculation

    :Params: a Fleurcalculation node
    :returns: Success, last result node, list with convergence behavior

    wf_parameters: {  'tria', 'nkpts', 'sigma', 'emin', 'emax'}
    defaults : tria = True, nkpts = 800, sigma=0.005, emin= -0.3, emax = 0.8
    """

    _workflowversion = '0.3.3'

    _default_options = {
        'resources': {
            'num_machines': 1
        },
        'max_wallclock_seconds': 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        # 'max_memory_kb' : None,
        'import_sys_environment': False,
        'environment_variables': {}
    }
    _default_wf_para = {
        'tria': True,
        'nkpts': 800,
        'sigma': 0.005,
        'emin': -0.30,
        'emax': 0.80,
        'add_comp_para': {
            'serial': False,
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        }
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('wf_parameters', valid_type=Dict, required=False, default=lambda: Dict(dict=cls._default_wf_para))
        spec.input('calc_parameters', valid_type=Dict, required=False)
        spec.input('settings', valid_type=Dict, required=False)
        spec.input('options', valid_type=Dict, required=False, default=lambda: Dict(dict=cls._default_options))
        spec.input('fleurinp', valid_type=FleurinpData, required=False)
        # TODO ggf run convergence first
        spec.input('remote_data', valid_type=RemoteData, required=False)
        #spec.input("inpgen", valid_type=Code, required=False)
        spec.input('fleur', valid_type=Code, required=True)
        spec.outline(cls.start, cls.create_new_fleurinp, cls.run_fleur, cls.return_results)
        # spec.dynamic_output()

    def start(self):
        '''
        check parameters, what condictions? complete?
        check input nodes
        '''
        # input check ### ? or done automaticly, how optional?
        # check if fleuinp corresponds to fleur_calc
        self.report('Started dos workflow version {}'
                    # "Workchain node identifiers: ")#{}"
                    ''.format(self._workflowversion))  # ProcessRegistry().current_calc_node))

        self.ctx.fleurinp1 = ''
        self.ctx.last_calc = None
        self.ctx.successful = False
        self.ctx.warnings = []

        wf_dict = self.inputs.wf_parameters.get_dict()

        # if MPI in code name, execute parallel
        self.ctx.serial = wf_dict.get('serial', False)

        # set values, or defaults
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)

        inputs = self.inputs
        if 'options' in inputs:
            self.ctx.options = inputs.options.get_dict()

        if 'remote_data' in inputs:
            self.ctx.remote = inputs.remote_data

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ('The code you provided for FLEUR does not use the plugin fleur.fleur')
                # self.control_end_wc(error)
                self.report(error)
                return 1

    def create_new_fleurinp(self):
        """
        create a new fleurinp from the old with certain parameters
        """
        # TODO allow change of kpoint mesh?, tria?
        wf_dict = self.inputs.wf_parameters.get_dict()
        nkpts = wf_dict.get('nkpts', 800)
        # how can the user say he want to use the given kpoint mesh, ZZ nkpts : False/0
        tria = wf_dict.get('tria', True)
        sigma = wf_dict.get('sigma', 0.005)
        emin = wf_dict.get('emin', -0.30)
        emax = wf_dict.get('emax', 0.80)

        fleurmode = FleurinpModifier(self.inputs.fleurinp)

        # change_dict = {'dos': True, 'ndir' : -1, 'minEnergy' : self.inputs.wf_parameters.get_dict().get('minEnergy', -0.30000000),
        # 'maxEnergy' :  self.inputs.wf_parameters.get_dict().get('manEnergy','0.80000000'),
        # 'sigma' :  self.inputs.wf_parameters.get_dict().get('sigma', '0.00500000')}
        change_dict = {'dos': True, 'ndir': -1, 'minEnergy': emin, 'maxEnergy': emax, 'sigma': sigma}

        fleurmode.set_inpchanges(change_dict)
        if tria:
            change_dict = {'mode': 'tria'}
            fleurmode.set_inpchanges(change_dict)
        if nkpts:
            fleurmode.set_nkpts(count=nkpts)
            # fleurinp_new.replace_tag()
        fleurmode.show(validate=True, display=False)  # needed?
        fleurinp_new = fleurmode.freeze()
        self.ctx.fleurinp1 = fleurinp_new
        # print(fleurinp_new)
        # print(fleurinp_new.folder.get_subfolder('path').get_abs_path(''))

    def run_fleur(self):
        """
        run a FLEUR calculation
        """
        fleurin = self.ctx.fleurinp1
        remote = self.inputs.remote_data
        code = self.inputs.fleur

        options = self.ctx.options

        inputs = get_inputs_fleur(code, remote, fleurin, options, add_comp_para=self.ctx.wf_dict['add_comp_para'])
        future = submit(FleurCalculation, **inputs)

        return ToContext(last_calc=future)  # calcs.append(future),

    def return_results(self):
        '''
        return the results of the calculations
        '''
        # TODO more here
        self.report('Dos workflow Done')
        # self.report('A DOS was calculated for calculation {} and is found under pk=, '
        #      'calculation {}')#.format(self.ctx.last_calc, ctx['last_calc'] )

        # check if dos file exists: if not succesful = False
        # TODO be careful with general DOS.X
        dosfilename = 'DOS.1'  # ['DOS.1', 'DOS.2', ...]
        # TODO this should be easier...
        dosfilepath = self.ctx.last_calc.get_outputs_dict()['retrieved'].folder.get_subfolder('path').get_abs_path(
            dosfilename)
        print(dosfilepath)
        # dosfilepath = "path to dosfile" # Array?
        if os.path.isfile(dosfilepath):
            self.ctx.successful = True
        else:
            dosfilepath = None
            self.report('!NO DOS.1 file was found, something went wrong!')

        outputnode_dict = {}

        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['Warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        #outputnode_dict['last_calc_pk'] = self.ctx.last_calc.pk
        #outputnode_dict['last_calc_uuid'] = self.ctx.last_calc.uuid
        outputnode_dict['dosfile'] = dosfilepath
        # add nkpoints, emin, emax, sigma, tria

        # print outputnode_dict
        outputnode = Dict(dict=outputnode_dict)
        outdict = {}
        # TODO parse dos to dosnode
        #dosnode = ''
        #outdict['output_band'] = dosnode
        # or if spin =2
        #outdict['output_band1'] = dosnode1
        #outdict['output_band2'] = dosnode2
        outdict['output_dos_wc_para'] = outputnode
        # print outdict
        for k, v in outdict.items():
            self.out(k, v)
